#! /usr/bin/env python

"""This module is home to a suite of MAST queries that gather bulk
properties of available JWST data for JWQL.

Authors
-------

    Joe Filippazzo

Use
---

    To get an inventory of all JWST files do:
    ::

        from jwql.jwql_monitors import monitor_mast
        inventory, keywords = monitor_mast.jwst_inventory()
"""

import logging
import os

from astroquery.mast import Mast
from bokeh.embed import components
from bokeh.io import save, output_file
import pandas as pd

from jwql.utils.constants import JWST_INSTRUMENT_NAMES, JWST_DATAPRODUCTS
from jwql.utils.logging_functions import configure_logging, log_info, log_fail
from jwql.utils.permissions import set_permissions
from jwql.utils.utils import get_config
from jwql.utils.plotting import bar_chart


def instrument_inventory(instrument, dataproduct=JWST_DATAPRODUCTS,
                         add_filters=None, add_requests=None,
                         caom=False, return_data=False):
    """Get the counts for a given instrument and data product

    Parameters
    ----------
    instrument: str
        The instrument name, i.e. one of ['niriss','nircam','nirspec',
        'miri','fgs']
    dataproduct: sequence, str
        The type of data product to search
    add_filters: dict
        The ('paramName':'values') pairs to include in the 'filters'
        argument of the request e.g. add_filters = {'filter':'GR150R'}
    add_requests: dict
        The ('request':'value') pairs to include in the request
        e.g. add_requests = {'pagesize':1, 'page':1}
    caom: bool
        Query CAOM service
    return_data: bool
        Return the actual data instead of counts only

    Returns
    -------
    int, dict
        The number of database records that satisfy the search criteria
        or a dictionary of the data if `return_data=True`
    """
    filters = []

    # Make sure the dataproduct is a list
    if isinstance(dataproduct, str):
        dataproduct = [dataproduct]

    # Make sure the instrument is supported
    if instrument.lower() not in [ins.lower() for ins in JWST_INSTRUMENT_NAMES]:
        raise TypeError('Supported instruments include:', JWST_INSTRUMENT_NAMES)

    # CAOM service
    if caom:

        # Declare the service
        service = 'Mast.Caom.Filtered'

        # Set the filters
        filters += [{'paramName': 'obs_collection', 'values': ['JWST']},
                    {'paramName': 'instrument_name', 'values': [instrument]},
                    {'paramName': 'dataproduct_type', 'values': dataproduct}]

    # Instruent filtered service
    else:

        # Declare the service
        service = 'Mast.Jwst.Filtered.{}'.format(instrument.title())

    # Include additonal filters
    if isinstance(add_filters, dict):
        filters += [{"paramName": name, "values": [val]}
                    for name, val in add_filters.items()]

    # Assemble the request
    params = {'columns': 'COUNT_BIG(*)',
              'filters': filters,
              'removenullcolumns': True}

    # Just get the counts
    if return_data:
        params['columns'] = '*'

    # Add requests
    if isinstance(add_requests, dict):
        params.update(add_requests)

    response = Mast.service_request_async(service, params)
    result = response[0].json()

    # Return all the data
    if return_data:
        return result

    # Or just the counts
    else:
        return result['data'][0]['Column1']


def instrument_keywords(instrument, caom=False):
    """Get the keywords for a given instrument service

    Parameters
    ----------
    instrument: str
        The instrument name, i.e. one of ['niriss','nircam','nirspec',
        'miri','fgs']
    caom: bool
        Query CAOM service

    Returns
    -------
    pd.DataFrame
        A DataFrame of the keywords
    """
    # Retrieve one dataset to get header keywords
    sample = instrument_inventory(instrument, return_data=True, caom=caom,
                                  add_requests={'pagesize': 1, 'page': 1})
    data = [[i['name'], i['type']] for i in sample['fields']]
    keywords = pd.DataFrame(data, columns=('keyword', 'dtype'))

    return keywords


def jwst_inventory(instruments=JWST_INSTRUMENT_NAMES,
                   dataproducts=['image', 'spectrum', 'cube'],
                   caom=False, plot=False):
    """Gather a full inventory of all JWST data in each instrument
    service by instrument/dtype

    Parameters
    ----------
    instruments: sequence
        The list of instruments to count
    dataproducts: sequence
        The types of dataproducts to count
    caom: bool
        Query CAOM service
    plot: bool
        Return a pie chart of the data

    Returns
    -------
    astropy.table.table.Table
        The table of record counts for each instrument and mode
    """
    logging.info('Searching database...')
    # Iterate through instruments
    inventory, keywords = [], {}
    for instrument in instruments:
        ins = [instrument]
        for dp in dataproducts:
            count = instrument_inventory(instrument, dataproduct=dp, caom=caom)
            ins.append(count)

        # Get the total
        ins.append(sum(ins[-3:]))

        # Add it to the list
        inventory.append(ins)

        # Add the keywords to the dict
        keywords[instrument] = instrument_keywords(instrument, caom=caom)

    logging.info('Completed database search for {} instruments and {} data products.'.
                 format(instruments, dataproducts))

    # Make the table
    all_cols = ['instrument'] + dataproducts + ['total']
    table = pd.DataFrame(inventory, columns=all_cols)

    # Plot it
    if plot:
        # Determine plot location and names
        output_dir = get_config()['outputs']

        if caom:
            output_filename = 'database_monitor_caom'
        else:
            output_filename = 'database_monitor_jwst'

        # Make the plot
        plt = bar_chart(table, 'instrument', dataproducts,
                        title="JWST Inventory")

        # Save the plot as full html
        html_filename = output_filename + '.html'
        outfile = os.path.join(output_dir, 'monitor_mast', html_filename)
        output_file(outfile)
        save(plt)
        set_permissions(outfile)

        logging.info('Saved Bokeh plots as HTML file: {}'.format(html_filename))

        # Save the plot as components
        plt.sizing_mode = 'stretch_both'
        script, div = components(plt)

        div_outfile = os.path.join(output_dir, 'monitor_mast', output_filename + "_component.html")
        with open(div_outfile, 'w') as f:
            f.write(div)
            f.close()
        set_permissions(div_outfile)

        script_outfile = os.path.join(output_dir, 'monitor_mast', output_filename + "_component.js")
        with open(script_outfile, 'w') as f:
            f.write(script)
            f.close()
        set_permissions(script_outfile)

        logging.info('Saved Bokeh components files: {}_component.html and {}_component.js'.format(
            output_filename, output_filename))

    # Melt the table
    table = pd.melt(table, id_vars=['instrument'],
                    value_vars=dataproducts,
                    value_name='files', var_name='dataproduct')

    return table, keywords


@log_fail
@log_info
def monitor_mast():
    """Tabulates the inventory of all JWST data products in the MAST
    archive and generates plots.
    """
    logging.info('Beginning database monitoring.')

    # Perform inventory of the JWST service
    jwst_inventory(instruments=JWST_INSTRUMENT_NAMES,
                   dataproducts=['image', 'spectrum', 'cube'],
                   caom=False, plot=True)

    # Perform inventory of the CAOM service
    jwst_inventory(instruments=JWST_INSTRUMENT_NAMES,
                   dataproducts=['image', 'spectrum', 'cube'],
                   caom=True, plot=True)


if __name__ == '__main__':

    # Configure logging
    module = os.path.basename(__file__).strip('.py')
    configure_logging(module)

    # Run the monitors
    monitor_mast()



"""Defines the forms for the ``jwql`` web app.

Django allows for an object-oriented model representation of forms for
users to provide input through HTTP POST methods. This module defines
all of the forms that are used across the various webpages used for the
JWQL application.

Authors
-------

    - Lauren Chambers
    - Johannes Sahlmann

Use
---

    This module is used within ``views.py`` as such:
    ::
        from .forms import FileSearchForm
        def view_function(request):
            form = FileSearchForm(request.POST or None)

            if request.method == 'POST':
                if form.is_valid():
                    # Process form input and redirect
                    return redirect(new_url)

            template = 'some_template.html'
            context = {'form': form, ...}
            return render(request, template, context)

References
----------
    For more information please see:
        ``https://docs.djangoproject.com/en/2.1/topics/forms/``

Dependencies
------------
    The user must have a configuration file named ``config.json``
    placed in the ``jwql/utils/`` directory.

"""
import glob
import os

from astropy.time import Time, TimeDelta
from django import forms
from django.shortcuts import redirect

from jwedb.edb_interface import is_valid_mnemonic
from jwql.utils.constants import JWST_INSTRUMENT_NAMES_SHORTHAND
from jwql.utils.utils import get_config, filename_parser

FILESYSTEM_DIR = os.path.join(get_config()['jwql_dir'], 'filesystem')


class FileSearchForm(forms.Form):
    """Single-field form to search for a proposal or fileroot."""

    # Define search field
    search = forms.CharField(label='', max_length=500, required=True,
                             empty_value='Search')

    # Initialize attributes
    fileroot_dict = None
    search_type = None
    instrument = None

    def clean_search(self):
        """Validate the "search" field.

        Check that the input is either a proposal or fileroot, and one
        that matches files in the filesystem.

        Returns
        -------
        str
            The cleaned data input into the "search" field

        """
        # Get the cleaned search data
        search = self.cleaned_data['search']

        # Make sure the search is either a proposal or fileroot
        if len(search) == 5 and search.isnumeric():
            self.search_type = 'proposal'
        elif self._search_is_fileroot(search):
            self.search_type = 'fileroot'
        else:
            raise forms.ValidationError('Invalid search term {}. Please provide proposal number '
                                        'or file root.'.format(search))

        # If they searched for a proposal...
        if self.search_type == 'proposal':
            # See if there are any matching proposals and, if so, what
            # instrument they are for
            search_string = os.path.join(FILESYSTEM_DIR, 'jw{}'.format(search),
                                         '*{}*.fits'.format(search))
            all_files = glob.glob(search_string)
            if len(all_files) > 0:
                all_instruments = []
                for file in all_files:
                    instrument = filename_parser(file)['instrument']
                    all_instruments.append(instrument)
                if len(set(all_instruments)) > 1:
                    raise forms.ValidationError('Cannot return result for proposal with multiple '
                                                'instruments.')

                self.instrument = all_instruments[0]
            else:
                raise forms.ValidationError('Proposal {} not in the filesystem.'.format(search))

        # If they searched for a fileroot...
        elif self.search_type == 'fileroot':
            # See if there are any matching fileroots and, if so, what
            # instrument they are for
            search_string = os.path.join(FILESYSTEM_DIR, search[:7], '{}*.fits'.format(search))
            all_files = glob.glob(search_string)

            if len(all_files) == 0:
                raise forms.ValidationError('Fileroot {} not in the filesystem.'.format(search))

            instrument = search.split('_')[-1][:3]
            self.instrument = JWST_INSTRUMENT_NAMES_SHORTHAND[instrument]

        return self.cleaned_data['search']

    def _search_is_fileroot(self, search):
        """Determine if a search value is formatted like a fileroot.

        Parameters
        ----------
        search : str
            The search term input by the user.

        Returns
        -------
        bool
            Is the search term formatted like a fileroot?

        """
        try:
            self.fileroot_dict = filename_parser(search)
            return True
        except ValueError:
            return False

    def redirect_to_files(self):
        """Determine where to redirect the web app based on user input.

        Returns
        -------
        HttpResponseRedirect object
            Outgoing redirect response sent to the webpage

        """
        # Process the data in form.cleaned_data as required
        search = self.cleaned_data['search']

        # If they searched for a proposal
        if self.search_type == 'proposal':
            return redirect('/{}/archive/{}'.format(self.instrument, search))

        # If they searched for a file root
        elif self.search_type == 'fileroot':
            return redirect('/{}/{}'.format(self.instrument, search))


class MnemonicSearchForm(forms.Form):
    """A single-field form to search for a mnemonic in the DMS EDB."""

    # Define search field
    search = forms.CharField(label='', max_length=500, required=True,
                             empty_value='Search', initial='SA_ZFGOUTFOV')

    # Initialize attributes
    search_type = None

    def clean_search(self):
        """Validate the "search" field.

        Check that the input is a valid mnemonic identifier.

        Returns
        -------
        str
            The cleaned data input into the "search" field

        """
        # Get the cleaned search data
        search = self.cleaned_data['search']

        # Make sure the search is a valid mnemonic identifier
        if is_valid_mnemonic(search):
            self.search_type = 'mnemonic'
        else:
            raise forms.ValidationError('Invalid search term {}. Please enter a valid DMS EDB '
                                        'mnemonic.'.format(search))

        return self.cleaned_data['search']


class MnemonicQueryForm(forms.Form):
    """A triple-field form to query mnemonic records in the DMS EDB."""

    production_mode = False

    if production_mode:
        # times for default query (one day one week ago)
        now = Time.now()
        delta_day = -7.
        range_day = 1.
        default_start_time = now + TimeDelta(delta_day, format='jd')
        default_end_time = now + TimeDelta(delta_day + range_day, format='jd')
    else:
        # example for testing
        default_start_time = Time('2019-01-16 00:00:00.000', format='iso')
        default_end_time = Time('2019-01-16 00:01:00.000', format='iso')

    default_mnemonic_identifier = 'IMIR_HK_ICE_SEC_VOLT4'

    # Define search fields
    search = forms.CharField(label='mnemonic', max_length=500, required=True,
                             initial=default_mnemonic_identifier, empty_value='Search',
                             help_text="Mnemonic identifier")

    start_time = forms.CharField(label='start', max_length=500, required=False,
                                 initial=default_start_time.iso, help_text="Start time")

    end_time = forms.CharField(label='end', max_length=500, required=False,
                               initial=default_end_time.iso, help_text="End time")

    # Initialize attributes
    search_type = None

    def clean_search(self):
        """Validate the "search" field.

        Check that the input is a valid mnemonic identifier.

        Returns
        -------
        str
            The cleaned data input into the "search" field

        """
        # Get the cleaned search data
        search = self.cleaned_data['search']

        if is_valid_mnemonic(search):
            self.search_type = 'mnemonic'
        else:
            raise forms.ValidationError('Invalid search term {}. Please enter a valid DMS EDB '
                                        'mnemonic.'.format(search))

        return self.cleaned_data['search']

    def clean_start_time(self):
        """Validate the start time.

        Returns
        -------
        str
           The cleaned data input into the start_time field

        """
        start_time = self.cleaned_data['start_time']
        try:
            Time(start_time, format='iso')
        except ValueError:
            raise forms.ValidationError('Invalid start time {}. Please enter a time in iso format, '
                                        'e.g. {}'.format(start_time, self.default_start_time))
        return self.cleaned_data['start_time']

    def clean_end_time(self):
        """Validate the end time.

        Returns
        -------
        str
           The cleaned data input into the end_time field

        """
        end_time = self.cleaned_data['end_time']
        try:
            Time(end_time, format='iso')
        except ValueError:
            raise forms.ValidationError('Invalid end time {}. Please enter a time in iso format, '
                                        'e.g. {}.'.format(end_time, self.default_end_time))

        if 'start_time' in self.cleaned_data.keys():
            # verify that end_time is later than start_time
            if self.cleaned_data['end_time'] <= self.cleaned_data['start_time']:
                raise forms.ValidationError('Invalid time inputs. End time is required to be after'
                                            ' Start time.')

        return self.cleaned_data['end_time']


class MnemonicExplorationForm(forms.Form):
    """A sextuple-field form to explore the EDB mnemonic inventory."""

    default_description = 'centroid data'

    # Define search fields
    description = forms.CharField(label='description', max_length=500, required=False,
                                  initial=default_description, help_text="Description")
    sql_data_type = forms.CharField(label='sqlDataType', max_length=500, required=False,
                                    help_text="sqlDataType")
    subsystem = forms.CharField(label='subsystem', max_length=500, required=False,
                                help_text="subsystem")
    tlm_identifier = forms.CharField(label='tlmIdentifier', max_length=500, required=False,
                                     help_text="Numerical ID (tlmIdentifier)")
    tlm_mnemonic = forms.CharField(label='tlmMnemonic', max_length=500, required=False,
                                   help_text="String ID (tlmMnemonic)")
    unit = forms.CharField(label='unit', max_length=500, required=False,
                           help_text="unit")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
mypy + typing demo for JWQL dev meeting 2018-1-3

Part 3: mypy for type checking

Many thanks to Tommy Tutone...
"""

from typing import NewType

#mypy can check for incorrectly-typed variables

bad_variable: str = 1 #no runtime error

#This can especially be useful when using pre-annotation, since types can be
#hinted before calculations, I/O, or other complex code determines its value.
jenny: str

#mypy can also check function arguments and return values
def ive_got_your_number(num: int) -> bool:
    if num == 867_5309:
        return True
    else:
        return "Jenny don't change your number"

ive_got_your_number("jenny") #no runtime error
ive_got_your_number(555_1212) #no runtime error

if ive_got_your_number(8675_309):
    jenny = 867_5309 #no runtime error
else:
    jenny = "Don't change your number"
    
#If for some reason you don't want a particular variable's type to be checked, 
#then use comment syntax and "ignore"
dummy = None # type: ignore # otherwise this will throw a mypy error!

#mypy can handle user-created types
UserID = NewType("UserID", str)

gray: UserID = UserID("Gray")
kanarek: UserID = "Kanarek" #no runtime error

user: UserID = gray + kanarek #no runtime error

def get_first_char(user: UserID) -> str:
    return user[0]

get_first_char(gray)
get_first_char("Gray") #no runtime error

    
#mypy can help you figure out the types of variables, if it's complicated to 
#find out beforehand
reveal_type(1)



#! /usr/bin/env python

"""Tests for the ``calculations`` module.

Authors
-------

    - Bryan Hilbert

Use
---

    These tests can be run via the command line (omit the ``-s`` to
    suppress verbose output to stdout):
    ::

        pytest -s test_calculations.py
"""

import numpy as np

from jwql.utils import calculations


def test_double_gaussian_fit():
    """Test the double Gaussian fitting function"""

    amplitude1 = 500
    mean_value1 = 0.5
    sigma_value1 = 0.05
    amplitude2 = 300
    mean_value2 = 0.4
    sigma_value2 = 0.03

    bin_centers = np.arange(0., 1.1, 0.007)
    input_params = [amplitude1, mean_value1, sigma_value1, amplitude2, mean_value2, sigma_value2]
    input_values = calculations.double_gaussian(bin_centers, *input_params)

    initial_params = [np.max(input_values), 0.55, 0.1, np.max(input_values), 0.5, 0.05]
    params, sigma = calculations.double_gaussian_fit(bin_centers, input_values, initial_params)

    assert np.allclose(np.array(params[0:3]), np.array([amplitude2, mean_value2, sigma_value2]),
                       atol=0, rtol=0.000001)
    assert np.allclose(np.array(params[3:]), np.array([amplitude1, mean_value1, sigma_value1]),
                       atol=0, rtol=0.000001)


def test_gaussian1d_fit():
    """Test histogram fitting function"""

    mean_value = 0.5
    sigma_value = 0.05
    image = np.random.normal(loc=mean_value, scale=sigma_value, size=(100, 100))
    hist, bin_edges = np.histogram(image, bins='auto')
    bin_centers = (bin_edges[1:] + bin_edges[0: -1]) / 2.
    initial_params = [np.max(hist), 0.55, 0.1]
    amplitude, peak, width = calculations.gaussian1d_fit(bin_centers, hist, initial_params)

    assert np.isclose(peak[0], mean_value, atol=0.0035, rtol=0.)
    assert np.isclose(width[0], sigma_value, atol=0.0035, rtol=0.)
    assert ((mean_value <= peak[0]+7*peak[1]) & (mean_value >= peak[0]-7*peak[1]))
    assert ((sigma_value <= width[0]+7*width[1]) & (sigma_value >= width[0]-7*width[1]))


def test_mean_image():
    """Test the sigma-clipped mean and stdev image calculator"""

    # Create a stack of 50 5x5 pixel images
    nstack = 50
    cube = np.zeros((nstack, 5, 5))

    # Set alternating frames equal to 4 and 5
    for i in range(nstack):
        if i % 2 == 0:
            cube[i, :, :] = 4.
        else:
            cube[i, :, :] = 5.

    # Insert a few signal values that will be removed by sigma clipping.
    # Make sure you "remove" and equal number of 4's and 5's from each
    # pixel in order to keep the mean at 4.5 and dev at 0.5
    cube[0, 0, 0] = 55.
    cube[1, 0, 0] = -78.
    cube[3, 3, 3] = 150.
    cube[2, 3, 3] = 32.
    cube[1, 4, 4] = -96.
    cube[4, 4, 4] = -25.
    mean_img, dev_img = calculations.mean_image(cube, sigma_threshold=3)

    assert np.all(mean_img == 4.5)
    assert np.all(dev_img == 0.5)


def test_mean_stdev():
    """Test calcualtion of the sigma-clipped mean from an image"""

    image = np.zeros((50, 50)) + 1.
    badx = [1, 4, 10, 14, 16, 20, 22, 25, 29, 30]
    bady = [13, 27, 43, 21, 1, 32, 25, 21, 9, 14]
    for x, y in zip(badx, bady):
        image[y, x] = 100.

    meanval, stdval = calculations.mean_stdev(image, sigma_threshold=3)
    assert meanval == 1.
    assert stdval == 0.



"""Maps URL paths to views in the ``jwql`` project.

This module connects requested URL paths to the corresponding view in
``views.py`` for each webpage in the JWQL application. When Django is
provided a path, it searches through the urlpatterns list provided
here until it finds one that matches. It then calls the assigned view
to load the appropriate webpage, passing an HttpRequest object.

Authors
-------

    - Lauren Chambers

Use
---

    Function views
        1. Add an import:
            from my_app import views
        2. Add a URL to urlpatterns:
            path('', views.home, name='home')
    Class-based views
        1. Add an import:
            from other_app.views import Home
        2. Add a URL to urlpatterns:
            path('', Home.as_view(), name='home')
    Including another URLconf
        1. Import the include() function:
            from django.urls import include, path
        2. Add a URL to urlpatterns:
            path('blog/', include('blog.urls'))

References
----------

    For more information please see:
        ``https://docs.djangoproject.com/en/2.0/topics/http/urls/``

Notes
-----

    Be aware that when a url is requested, it will be directed to the
    first matching path in the ``urlpatterns`` list that it finds. The
    ``<str:var>`` tag is just a placeholder. To avoid complications,
    users should order their paths in order from shortest to longest,
    and after that from most to least specific.
"""

from django.contrib import admin
from django.urls import include, path

from ..apps.jwql import views

# Define custom error page views
handler404 = views.not_found  # Page not found
handler500 = views.not_found  # Internal error
handler403 = views.not_found  # Permission denied
handler400 = views.not_found  # Bad request

urlpatterns = [
    path('', include('jwql.website.apps.jwql.urls')),
    path('admin/', admin.site.urls),
]



#! /usr/bin/env python

"""This module monitors the status of the ``jwql`` monitors via their
log files. Basic results (e.g. ``success``, ``failure``) are collected
and placed in a ``bokeh`` table for display on the web app.

Authors
-------

    - Bryan Hilbert

Use
---

    This module can be executed as such:

    ::

        from jwql.jwql_monitors import monitor_cron_jobs
        monitor_cron_jobs.status()

Dependencies
------------

    The user must have a configuration file named ``config.json``
    placed in the ``utils`` directory.
"""

from datetime import datetime
import logging
import numpy as np
import os
import time

from bokeh.io import save, output_file
from bokeh.models import ColumnDataSource
from bokeh.models.widgets import DataTable, DateFormatter, HTMLTemplateFormatter, TableColumn

from jwql.utils.logging_functions import configure_logging, log_info, log_fail
from jwql.utils.permissions import set_permissions
from jwql.utils.utils import get_config


def create_table(status_dict):
    """Create interactive ``bokeh`` table containing the logfile status
    results.

    Parameters
    ----------
    status_dict : dict
        Nested dictionary with status results from all logfiles
    """
    # Rearrange the nested dictionary into a non-nested dict for the table
    filenames = []
    dates = []
    missings = []
    results = []
    for key in status_dict:
        filenames.append(status_dict[key]['logname'])
        dates.append(datetime.fromtimestamp(status_dict[key]['latest_time']))
        missings.append(str(status_dict[key]['missing_file']))
        results.append(status_dict[key]['status'])

    # div to color the boxes in the status column
    success_template = """
    <div style="background:<%=
        (function colorfromstr(){
            if(value == "success"){
                return("green")}
            else{return("red")}
            }()) %>;
        color: white">
    <%= value %></div>
    """

    # div to color the boxes in the column for possibly late logfiles
    missing_template = """
    <div style="background:<%=
        (function colorfrombool(){
            if(value == "True"){
                return("orange")}
            else{return("green")}
            }()) %>;
        color: white">
    <%= value %></div>
    """
    success_formatter = HTMLTemplateFormatter(template=success_template)
    missing_formatter = HTMLTemplateFormatter(template=missing_template)

    data = dict(name=list(status_dict.keys()), filename=filenames, date=dates, missing=missings,
                result=results)
    source = ColumnDataSource(data)

    datefmt = DateFormatter(format="RFC-2822")
    columns = [
        TableColumn(field="name", title="Monitor Name", width=200),
        TableColumn(field="filename", title="Most Recent File", width=350),
        TableColumn(field="date", title="Most Recent Time", width=200, formatter=datefmt),
        TableColumn(field="missing", title="Possible Missing File", width=200, formatter=missing_formatter),
        TableColumn(field="result", title="Status", width=100, formatter=success_formatter),
    ]
    data_table = DataTable(source=source, columns=columns, width=800, height=280, index_position=None)

    # Get output directory for saving the table files
    output_dir = get_config()['outputs']
    output_filename = 'cron_status_table'

    # Save full html
    html_outfile = os.path.join(output_dir, 'monitor_cron_jobs', '{}.html'.format(output_filename))
    output_file(html_outfile)
    save(data_table)
    try:
        set_permissions(html_outfile)
    except PermissionError:
        logging.warning('Unable to set permissions for {}'.format(html_outfile))
    logging.info('Saved Bokeh full HTML file: {}'.format(html_outfile))


def find_latest(logfiles):
    """Given a list of log files in a directory, identify the most
    recent. The way that ``jwql.utils.logging_functions.make_log_file``
    is set up, log files for all monitors are guaranteed to be the name
    of the monitor followed by the datetime that they were run, so we
    should be able to simply sort the filenames and the last will be the
    most recent.

    Parameters
    ----------
    logfiles : list
        List of logfiles in the directory

    Returns
    -------
    latest : str
        Filename of the most recent file

    latest_time : float
        Time associated with the most recent log file
    """
    latest = sorted(logfiles)[-1]
    latest_time = os.path.getctime(latest)
    return (latest, latest_time)


def get_cadence(filenames):
    """Calculate the cadence of the log files in a given directory.
    Use timestamps

    Parameters
    ----------
    filenames : list
        List of log files to examine

    Returns
    -------
    mean_delta : float
        Mean time in seconds between the appearance of consecutive log
        files

    stdev_delta : float
        Standard deviation in seconds between the appearance of
        consecutive log files
    """
    minimum_log_num = 3  # Set to a low value for now since we don't have many logfiles
    times = [os.path.getctime(filename) for filename in filenames]
    if len(times) >= minimum_log_num:
        sorted_times = np.array(sorted(times))
        delta_times = sorted_times[1:] - sorted_times[0:-1]
        mean_delta = np.mean(delta_times)
        stdev_delta = np.std(delta_times)
    else:
        # If there are < minimum_log_num logfiles, then let's assume we can't
        # get a reliable measure of cadence. Fall back to a value of
        # 1 year between files, to avoid accidentally flagging this monitor
        # as running late in the subsequent check
        mean_delta = 31556736.0  # sec per year
        stdev_delta = 31556736.0  # sec per year
    return mean_delta, stdev_delta


def missing_file_check(avg_time_between, uncertainty, latest_file):
    """Given the name of the most recent log file, along with the
    historical average time between files and the stdev of the time
    between files, determine whether we expect a more recent log file
    than the file given. This could hint at a problem with the cron job
    used to create the log files.

    Parameters
    ----------
    avg_time_between : float
        Average number of seconds between log files

    uncertainty : float
        Standard deviation of the number of seconds between log files

    latest_file : str
        Name of the most recent log file

    Returns
    -------
    late : bool
        True = We expect a more recent file than that given
        False =  It is reasonable that the file given is the most
        recent
    """
    latest_time = os.path.getctime(latest_file)
    now = time.time()
    time_since_latest = now - latest_time
    if time_since_latest > (avg_time_between + 3 * uncertainty):
        late = True
    else:
        late = False
    return late


@log_fail
@log_info
def status(production_mode=True):
    """Main function: determine the status of the instrument montiors
    by examining log files.

    Parameters
    ----------
    production_mode : bool
        If ``True``, look in the main log directory. If ``False``, look
        in the ``dev`` log file directory.

    Returns
    -------
    logfile_status : dict
        Nested dictionary containing the status for all monitors. Top
        level keys include all monitors. Within a given monitor, the
        value is a dictionary containing 'missing_file' and 'status'
        keys. 'missing_file' is a boolean describing whether or not
        there is a suspected missing log file based on the timestamps
        of the existing files. 'status' is a string that is either
        'success' or 'failure'.
    """
    # Begin logging
    logging.info("Beginning cron job status monitor")

    # Get main logfile path
    log_path = get_config()['log_dir']

    # If we are in development mode, the log files are in a slightly
    # different location than in production mode
    if production_mode:
        log_path = os.path.join(log_path, 'prod')
    else:
        log_path = os.path.join(log_path, 'dev')

    # Set up a dictionary to keep track of results
    logfile_status = {}

    # Get a list of the directories under the main logging directory.
    generator = os.walk(log_path, topdown=True)

    # Loop over monitors
    for subdir, subsubdir, filenames in generator:
        # When running in production mode, skip the 'dev' subdirectory,
        # as it contains the development version of the monitor logs
        if production_mode:
            subsubdir[:] = [dirname for dirname in subsubdir if dirname != 'dev']

        if len(filenames) > 0:
            monitor_name = subdir.split('/')[-1]

            # Avoid monitor_cron_jobs itseft
            if monitor_name != 'monitor_cron_jobs':

                log_file_list = [os.path.join(subdir, filename) for filename in filenames]

                # Find the cadence of the monitor
                delta_time, stdev_time = get_cadence(log_file_list)

                # Identify the most recent log file
                latest_log, latest_log_time = find_latest(log_file_list)

                # Check to see if we expect a file more recent than the latest
                missing_file = missing_file_check(delta_time, stdev_time, latest_log)
                if missing_file:
                    logging.warning('Expected a more recent {} logfile than {}'
                                    .format(monitor_name, os.path.basename(latest_log)))

                # Check the file for success/failure
                result = success_check(latest_log)
                logging.info('{}: Latest log file indicates {}'.format(monitor_name, result))

                # Add results to the dictionary
                logfile_status[monitor_name] = {'logname': os.path.basename(latest_log),
                                                'latest_time': latest_log_time,
                                                'missing_file': missing_file, 'status': result}

    # Create table of results using Bokeh
    create_table(logfile_status)
    logging.info('Cron job status monitor completed successfully.')


def success_check(filename):
    """Parse the given log file and check whether the script execution
    was successful or not

    Parameters
    ----------
    filename : str
        Name of the log file to parse

    Returns
    -------
    execution : str
        ``success`` or ``failure``
    """
    with open(filename, 'r') as file_obj:
        all_lines = file_obj.readlines()
    final_line = all_lines[-1]
    if 'complete' in final_line.lower():
        execution = 'success'
    else:
        execution = 'failure'
    return execution


if __name__ == '__main__':

    module = os.path.basename(__file__).strip('.py')
    configure_logging(module)

    status()



"""Defines the models for the ``jwql`` app.

** CURRENTLY NOT IN USE **

In Django, "a model is the single, definitive source of information
about your data. It contains the essential fields and behaviors of the
data you’re storing. Generally, each model maps to a single database
table" (from Django documentation). Each model contains fields, such
as character fields or date/time fields, that function like columns in
a data table. This module defines models that might be used to store
data related to the JWQL webpage. Interacts with the database located
at jwql/website/db.sqlite3.

Authors
-------
    - Lauren Chambers

Use
---
    This module is used as such:

    ::
        from models import MyModel
        data = MyModel.objects.filter(name="JWQL")

References
----------
    For more information please see:
        ```https://docs.djangoproject.com/en/2.0/topics/db/models/```
"""

import os

from django.db import models


INSTRUMENT_LIST = (('FGS', 'FGS'),
                   ('MIRI', 'MIRI'),
                   ('NIRCam', 'NIRCam'),
                   ('NIRISS', 'NIRISS'),
                   ('NIRSpec', 'NIRSpec'))

class BaseModel(models.Model):
    """A base model that other classes will inherit. Created to avoid
    an obscure error about a missing ``app_label``.
    """

    class Meta:
        abstract = True  # specify this model as an Abstract Model
        app_label = 'jwql'


class ImageData(BaseModel):
    """A model that collects image filepaths, instrument labels, and
    publishing date/time. Just an example used for learning django.

    Attributes
    ----------
    filepath : FilePathField object
        The full filepath of the datum
    inst : CharField object
        Name of the corresponding JWST instrument
    pub_date : FilePathField object
        Date and time when datum was added to the database.
    """

    inst = models.CharField('instrument', max_length=6, choices=INSTRUMENT_LIST, default=None)
    pub_date = models.DateTimeField('date published')
    filepath = models.FilePathField(path='/user/lchambers/jwql/')  #upload_to=str(inst))

    def filename(self):
        return os.path.basename(self.filepath)

    def __str__(self):
        return self.filename()

    class Meta:
        verbose_name_plural = "image data"
        db_table = 'imagedata'



#! /usr/bin/env python
"""Module generates conditions over one or more mnemonics

The modules purpose is to return True/False for any times by reference of
certain conditions. If for instance the condition "x>1" over a defined period of
time is needed, the module looks for all elements where the condition applies
and where it does not apply. This generates two lists, which contain the "start"
and "end" times of the condition.
A futher function combines the start- and endtimes to time-tuples between which
the condition is known as TRUE. A "state" function returns True/False for an
exact time attribute, whereby the condition is represented in binary form.

Authors
-------
    - Daniel Kühbacher

Use
---
    This module is not prepared for standalone use.

    For use in programm set condition up like below:

    import the module as follow:
    >>>import condition as cond

    generate list with required conditions:
    >>>con_set = [ cond.equal(m.mnemonic('IMIR_HK_POM_LOOP'),'OFF'),
                cond.smaller(m.mnemonic('IMIR_HK_ICE_SEC_VOLT1'),1),
                cond.greater(m.mnemonic('SE_ZIMIRICEA'),0.2)]

    generate object of condition with the con_set as attribute:
    >>>condition_object=cond.condition(con_set)

    Now the condition_object can return a True/False statement wheather
    the time given as attribut meets the conditions:

    >>>if condition.state(float(element['Primary Time'])):
        -> True when condition for the given time applies
        -> False when condition for the given time is not applicable

Dependencies
------------
    no external files needed

References
----------

Notes
-----

"""


class condition:
    """Class to hold several subconditions"""

    #contains list of representative time pairs for each subcondition
    cond_time_pairs = []
    #state of the condition
    __state = False

    #initializes condition through condition set
    def __init__(self, cond_set):
        """Initialize object with set of conditions
        Parameters
        ----------
        cond_set : list
            list contains subconditions objects
        """
        self.cond_set = cond_set

    #destructor -> take care that all time_pairs are deleted!
    def __del__(self):
        """Delete object - destructor method"""
        del self.cond_time_pairs[:]

    #prints all stored time pairs (for developement only)
    def print_times(self):
        """Print conditions time pairs on command line (developement)"""
        print('Available time pairs:')
        for times in self.cond_time_pairs:
            print('list: '+str(times))

    #returns a interval if time is anywhere in between
    def get_interval(self, time):
        """Returns time interval if availlable, where "time" is in between
        Parameters
        ----------
        time : float
            given time attribute
        Return
        ------
        time_pair : tuple
            pair of start_time and end_time where time is in between
        """
        end_time = 10000000
        start_time = 0

        #do for every condition
        for cond in self.cond_time_pairs:
            #do for every time pair in condition
            for pair in cond:
                if (time > pair[0]) and (time < pair[1]):
                    if (end_time > pair[1]) and (start_time < pair[0]):
                        start_time = pair[0]
                        end_time = pair[1]
                        break
                    else:
                        break

        if (end_time != 10000000) and (start_time != 0):
            return [start_time, end_time]
        else:
            return None


    #generates time pairs out of start and end times
    def generate_time_pairs(start_times, end_times):
        """Forms time pairs out of start times and end times
        Parameters
        ----------
        start_times : list
            contains all times where a condition applies
        end_times : list
            contains all times where the condition does not apply
        Return
        ------
        time_pair : list
            list of touples with start and end time
        """
        #internal use only
        time_pair: float = []

        #when the conditons doesn´t apply anyway
        if not start_times:
            time_pair.append((0,0))

        #check if the condition indicates an open time range
        elif not end_times:
            time_pair.append((start_times[0], 0))

        #generate time pairs
        #for each start time a higher or equal end time is searched for
        #these times form am touple which is appended to  time_pair : list
        else:
            time_hook = 0
            last_start_time = 0

            for start in list(sorted(set(start_times))):

                if(start > time_hook):
                    for end in list(sorted(set(end_times))):

                        if end > start:

                            time_pair.append((start, end))
                            time_hook = end
                            break

            if list(sorted(set(start_times)))[-1] > list(sorted(set(end_times)))[-1]:
                time_pair.append((list(sorted(set(end_times)))[-1], 0))

        return(time_pair)


    #returns state of the condition at a given time
    #if state(given time)==True -> condition is true
    #if state(given time)==False -> condition is false
    def state(self, time):
        """Checks whether condition is true of false at a given time
        Parameters
        ----------
        time : float
            input time for condition query
        Return
        ------
        state : bool
            True/False statement whether the condition applies or not
        """
        #checks condition for every sub condition in condition set (subconditions)

        state = self.__state

        for cond in self.cond_time_pairs:

            if self.__check_subcondition(cond, time):
                state = True
            else:
                state = False
                break

        return state


    def __check_subcondition(self, cond, time):

        #if there are no values availlable
        if cond[0][0] == 0:
            return False

        for time_pair in cond:
            #if just a positive time is availlable, return true
            if (time_pair[1] == 0) and (time > time_pair[0]):

                return True

            #if given time occurs between a time pair, return true
            elif (time_pair[0]) <= time and (time < time_pair[1]):

                return True

            else:
                pass


class equal(condition):
    """Class to hold single "is equal" subcondition"""

    stringval = True

    #add attributes to function - start function "cond_time_pairs()"
    def __init__(self, mnemonic, value, stringval=True):
        """Initializes subconditon
        Parameters
        ----------
        mnemonic : astropy table
            includes mnemomic engineering data and corresponding primary time
        value : str
            coparison value for equal statement
        """
        self.mnemonic = mnemonic
        self.value = value
        self.stringval = stringval
        condition.cond_time_pairs.append((self.cond_true_time()))


    #generates a list of time-touples (start_time, end_time) that mark the beginning and end of
    #wheather the condition is true or not
    def cond_true_time(self):
        """Filters all values that are equal to a given comparison value
        if equal: Primary time -> temp_start
        if not equal: Primary time -> temp_end
        Return
        ------
        time_p : list
            list of touples with start and end time
        """
        temp_start = []
        temp_end = []

        for key in self.mnemonic:

            #find all times whoses Raw values equal the given value
            if self.stringval:
                if key['value'] == self.value:
                    temp_start.append(key["time"])

                #find all end values
                else:
                    temp_end.append(key["time"])
            else:
                # just another option to compare float values
                if float(key['value']) == self.value:
                    temp_start.append(key["time"])

                #find all end values
                else:
                    temp_end.append(key["time"])

        time_p = condition.generate_time_pairs(temp_start, temp_end)
        return time_p

class unequal(condition):
    """Class to hold single "is unequal" subcondition"""

    #add attributes to function - start function "cond_time_pairs()"
    def __init__(self, mnemonic, value):
        """Initializes subconditon
        Parameters
        ----------
        mnemonic : astropy table
            includes mnemomic engineering data and corresponding primary time
        value : str
            coparison value for equal statement
        """
        self.mnemonic = mnemonic
        self.value = value
        condition.cond_time_pairs.append((self.cond_true_time()))


    #generates a list of time-touples (start_time, end_time) that mark the beginning and end of
    #wheather the condition is true or not
    def cond_true_time(self):
        """Filters all values that are equal to a given comparison value
        if equal: Primary time -> temp_start
        if not equal: Primary time -> temp_end
        Return
        ------
        time_p : list
            list of touples with start and end time
        """
        temp_start = []
        temp_end = []

        for key in self.mnemonic:

            #find all times whoses Raw values equal the given value
            if key['value'] != self.value:
                temp_start.append(key["time"])

            #find all end values
            else:
                temp_end.append(key["time"])

        time_p = condition.generate_time_pairs(temp_start, temp_end)
        return time_p

class greater(condition):
    """Class to hold single "greater than" subcondition"""

    #add attributes to function - start function "cond_time_pairs()"
    def __init__(self, mnemonic, value):
        """Initializes subconditon
        Parameters
        ----------
        mnemonic : astropy table
            includes mnemomic engineering data and corresponding primary time
        value : str
            coparison value for equal statement
        """
        self.mnemonic= mnemonic
        self.value=value
        condition.cond_time_pairs.append((self.cond_true_time()))

    def cond_true_time(self):
        """Filters all values that are greater than a given comparison value
        if equal: Primary time -> temp_start
        if not equal: Primary time -> temp_end
        Return
        ------
        time_p : list
            list of touples with start and end time
        """
        temp_start: float = []
        temp_end: float = []

        for key in self.mnemonic:

            #find all times whose Raw values are grater than the given value
            if float(key['value']) > self.value:
                temp_start.append(key["time"])

            #find all end values
            else:
                temp_end.append(key["time"])

        time_p = condition.generate_time_pairs(temp_start, temp_end)
        return time_p


class smaller(condition):
    """Class to hold single "greater than" subcondition"""

    #add attributes to function - start function "cond_time_pairs()"
    def __init__(self, mnemonic, value):
        """Initializes subconditon
        Parameters
        ----------
        mnemonic : astropy table
            includes mnemomic engineering data and corresponding primary time
        value : str
            coparison value for equal statement
        """
        self.mnemonic=mnemonic
        self.value=value
        condition.cond_time_pairs.append((self.cond_true_time()))

    def cond_true_time(self):
        """Filters all values that are greater than a given comparison value
        if equal: Primary time -> temp_start
        if not equal: Primary time -> temp_end
        Return
        ------
        time_p : list
            list of touples with start and end time
        """
        temp_start: float = []
        temp_end: float = []

        for key in self.mnemonic:

            #find all times whose Raw values are grater than the given value
            if float(key['value']) < self.value:
                temp_start.append(key["time"])

            #find all end values
            else:
                temp_end.append(key["time"])

        time_p = condition.generate_time_pairs(temp_start, temp_end)
        return time_p


if __name__ =='__main__':
    pass



"""Various utility functions for the ``jwql`` project.

Authors
-------

    - Matthew Bourque
    - Lauren Chambers

Use
---

    This module can be imported as such:

    >>> import utils
    settings = get_config()

References
----------

    Filename parser modified from Joe Hunkeler:
    https://gist.github.com/jhunkeler/f08783ca2da7bfd1f8e9ee1d207da5ff

    Various documentation related to JWST filename conventions:
    - https://jwst-docs.stsci.edu/display/JDAT/File+Naming+Conventions+and+Data+Products
    - https://innerspace.stsci.edu/pages/viewpage.action?pageId=94092600
    - https://innerspace.stsci.edu/pages/viewpage.action?spaceKey=SCSB&title=JWST+Science+Data+Products
    - https://jwst-docs.stsci.edu/display/JDAT/Understanding+Associations?q=association%20candidate
    - https://jwst-pipeline.readthedocs.io/en/stable/jwst/introduction.html#pipeline-step-suffix-definitions
    - JWST TR JWST-STScI-004800, SM-12
 """

import datetime
import getpass
import json
import os
import re
import shutil

from jwql.utils import permissions
from jwql.utils.constants import FILE_SUFFIX_TYPES, JWST_INSTRUMENT_NAMES_SHORTHAND

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


def copy_files(files, out_dir):
    """Copy a given file to a given directory. Only try to copy the file
    if it is not already present in the output directory.

    Parameters
    ----------
    files : list
        List of files to be copied

    out_dir : str
        Destination directory

    Returns
    -------
    success : list
        Files successfully copied (or that already existed in out_dir)

    failed : list
        Files that were not copied
    """

    # Copy files if they do not already exist
    success = []
    failed = []
    for input_file in files:
        input_new_path = os.path.join(out_dir, os.path.basename(input_file))
        if os.path.isfile(input_new_path):
            success.append(input_new_path)
        else:
            try:
                shutil.copy2(input_file, out_dir)
                success.append(input_new_path)
                permissions.set_permissions(input_new_path)
            except:
                failed.append(input_file)
    return success, failed


def download_mast_data(query_results, output_dir):
    """Example function for downloading MAST query results. From MAST
    website (``https://mast.stsci.edu/api/v0/pyex.html``)

    Parameters
    ----------
    query_results : list
        List of dictionaries returned by a MAST query.

    output_dir : str
        Directory into which the files will be downlaoded
    """

    # Set up the https connection
    server = 'mast.stsci.edu'
    conn = httplib.HTTPSConnection(server)

    # Dowload the products
    print('Number of query results: {}'.format(len(query_results)))

    for i in range(len(query_results)):

        # Make full output file path
        output_file = os.path.join(output_dir, query_results[i]['filename'])

        print('Output file is {}'.format(output_file))

        # Download the data
        uri = query_results[i]['dataURI']

        print('uri is {}'.format(uri))

        conn.request("GET", "/api/v0/download/file?uri="+uri)
        resp = conn.getresponse()
        file_content = resp.read()

        # Save to file
        with open(output_file, 'wb') as file_obj:
            file_obj.write(file_content)

        # Check for file
        if not os.path.isfile(output_file):
            print("ERROR: {} failed to download.".format(output_file))
        else:
            statinfo = os.stat(output_file)
            if statinfo.st_size > 0:
                print("DOWNLOAD COMPLETE: ", output_file)
            else:
                print("ERROR: {} file is empty.".format(output_file))
    conn.close()


def ensure_dir_exists(fullpath):
    """Creates dirs from ``fullpath`` if they do not already exist."""
    if not os.path.exists(fullpath):
        os.makedirs(fullpath)
        permissions.set_permissions(fullpath)


def filename_parser(filename):
    """Return a dictionary that contains the properties of a given
    JWST file (e.g. program ID, visit number, detector, etc.).

    Parameters
    ----------
    filename : str
        Path or name of JWST file to parse

    Returns
    -------
    filename_dict : dict
        Collection of file properties

    Raises
    ------
    ValueError
        When the provided file does not follow naming conventions
    """

    filename = os.path.basename(filename)
    file_root_name = (len(filename.split('.')) < 2)

    # Stage 1 and 2 filenames
    # e.g. "jw80500012009_01101_00012_nrcalong_uncal.fits"
    stage_1_and_2 = \
        r"jw" \
        r"(?P<program_id>\d{5})"\
        r"(?P<observation>\d{3})"\
        r"(?P<visit>\d{3})"\
        r"_(?P<visit_group>\d{2})"\
        r"(?P<parallel_seq_id>\d{1})"\
        r"(?P<activity>\w{2})"\
        r"_(?P<exposure_id>\d+)"\
        r"_(?P<detector>((?!_)[\w])+)"

    # Stage 2c outlier detection filenames
    # e.g. "jw94015002002_02108_00001_mirimage_o002_crf.fits"
    stage_2c = \
        r"jw" \
        r"(?P<program_id>\d{5})" \
        r"(?P<observation>\d{3})" \
        r"(?P<visit>\d{3})" \
        r"_(?P<visit_group>\d{2})" \
        r"(?P<parallel_seq_id>\d{1})" \
        r"(?P<activity>\w{2})" \
        r"_(?P<exposure_id>\d+)" \
        r"_(?P<detector>((?!_)[\w])+)"\
        r"_(?P<ac_id>(o\d{3}|(c|a|r)\d{4}))"

    # Stage 3 filenames with target ID
    # e.g. "jw80600-o009_t001_miri_f1130w_i2d.fits"
    stage_3_target_id = \
        r"jw" \
        r"(?P<program_id>\d{5})"\
        r"-(?P<ac_id>(o\d{3}|(c|a|r)\d{4}))"\
        r"_(?P<target_id>(t)\d{3})"\
        r"_(?P<instrument>(nircam|niriss|nirspec|miri|fgs))"\
        r"_(?P<optical_elements>((?!_)[\w-])+)"

    # Stage 3 filenames with source ID
    # e.g. "jw80600-o009_s00001_miri_f1130w_i2d.fits"
    stage_3_source_id = \
        r"jw" \
        r"(?P<program_id>\d{5})"\
        r"-(?P<ac_id>(o\d{3}|(c|a|r)\d{4}))"\
        r"_(?P<source_id>(s)\d{5})"\
        r"_(?P<instrument>(nircam|niriss|nirspec|miri|fgs))"\
        r"_(?P<optical_elements>((?!_)[\w-])+)"

    # Stage 3 filenames with target ID and epoch
    # e.g. "jw80600-o009_t001-epoch1_miri_f1130w_i2d.fits"
    stage_3_target_id_epoch = \
        r"jw" \
        r"(?P<program_id>\d{5})"\
        r"-(?P<ac_id>(o\d{3}|(c|a|r)\d{4}))"\
        r"_(?P<target_id>(t)\d{3})"\
        r"-epoch(?P<epoch>\d{1})"\
        r"_(?P<instrument>(nircam|niriss|nirspec|miri|fgs))"\
        r"_(?P<optical_elements>((?!_)[\w-])+)"

    # Stage 3 filenames with source ID and epoch
    # e.g. "jw80600-o009_s00001-epoch1_miri_f1130w_i2d.fits"
    stage_3_source_id_epoch = \
        r"jw" \
        r"(?P<program_id>\d{5})"\
        r"-(?P<ac_id>(o\d{3}|(c|a|r)\d{4}))"\
        r"_(?P<source_id>(s)\d{5})"\
        r"-epoch(?P<epoch>\d{1})"\
        r"_(?P<instrument>(nircam|niriss|nirspec|miri|fgs))"\
        r"_(?P<optical_elements>((?!_)[\w-])+)"

    # Time series filenames
    # e.g. "jw00733003001_02101_00002-seg001_nrs1_rate.fits"
    time_series = \
        r"jw" \
        r"(?P<program_id>\d{5})"\
        r"(?P<observation>\d{3})"\
        r"(?P<visit>\d{3})"\
        r"_(?P<visit_group>\d{2})"\
        r"(?P<parallel_seq_id>\d{1})"\
        r"(?P<activity>\w{2})"\
        r"_(?P<exposure_id>\d+)"\
        r"-seg(?P<segment>\d{3})"\
        r"_(?P<detector>\w+)"

    # Guider filenames
    # e.g. "jw00729011001_gs-id_1_image_cal.fits" or
    # "jw00799003001_gs-acq1_2019154181705_stream.fits"
    guider = \
        r"jw" \
        r"(?P<program_id>\d{5})" \
        r"(?P<observation>\d{3})" \
        r"(?P<visit>\d{3})" \
        r"_gs-(?P<guider_mode>(id|acq1|acq2|track|fg))" \
        r"_((?P<date_time>\d{13})|(?P<guide_star_attempt_id>\d{1}))"

    # Build list of filename types
    filename_types = [
        stage_1_and_2,
        stage_2c,
        stage_3_target_id,
        stage_3_source_id,
        stage_3_target_id_epoch,
        stage_3_source_id_epoch,
        time_series,
        guider]

    filename_type_names = [
        'stage_1_and_2',
        'stage_2c',
        'stage_3_target_id',
        'stage_3_source_id',
        'stage_3_target_id_epoch',
        'stage_3_source_id_epoch',
        'time_series',
        'guider'
    ]

    # Try to parse the filename
    for filename_type, filename_type_name in zip(filename_types, filename_type_names):

        # If full filename, try using suffix
        if not file_root_name:
            filename_type += r"_(?P<suffix>{}).*".format('|'.join(FILE_SUFFIX_TYPES))
        # If not, make sure the provided regex matches the entire filename root
        else:
            filename_type += r"$"

        elements = re.compile(filename_type)
        jwst_file = elements.match(filename)

        # Stop when you find a format that matches
        if jwst_file is not None:
            name_match = filename_type_name
            break

    try:
        # Convert the regex match to a dictionary
        filename_dict = jwst_file.groupdict()

        # Add the filename type to that dict
        filename_dict['filename_type'] = name_match

        # Also, add the instrument if not already there
        if 'instrument' not in filename_dict.keys():
            if name_match == 'guider':
                filename_dict['instrument'] = 'fgs'
            elif 'detector' in filename_dict.keys():
                filename_dict['instrument'] = JWST_INSTRUMENT_NAMES_SHORTHAND[
                    filename_dict['detector'][:3]
                ]

    # Raise error if unable to parse the filename
    except AttributeError:
        jdox_url = 'https://jwst-docs.stsci.edu/display/JDAT/' \
                   'File+Naming+Conventions+and+Data+Products'
        raise ValueError('Provided file {} does not follow JWST naming conventions.  See {} for further information.'.format(filename, jdox_url))

    return filename_dict


def filesystem_path(filename):
    """Return the full path to a given file in the filesystem

    Parameters
    ----------
    filename : str
        File to locate (e.g. ``jw86600006001_02101_00008_guider1_cal.fits``)

    Returns
    -------
    full_path : str
        Full path to the given file, including filename
    """

    filesystem_base = get_config()["filesystem"]

    # Subdirectory name is based on the proposal ID
    subdir = 'jw{}'.format(filename_parser(filename)['program_id'])
    full_path = os.path.join(filesystem_base, subdir, filename)

    # Check to see if the file exists
    if os.path.isfile(full_path):
        return full_path
    else:
        raise FileNotFoundError(('{} is not in the predicted location: {}'.format(filename, full_path)))


def get_base_url():
    """Return the beginning part of the URL to the ``jwql`` web app
    based on which user is running the software.

    If the admin account is running the code, the ``base_url`` is
    assumed to be the production URL.  If not, the ``base_url`` is
    assumed to be local.

    Returns
    -------
    base_url : str
        The beginning part of the URL to the ``jwql`` web app
    """

    username = getpass.getuser()
    if username == get_config()['admin_account']:
        base_url = 'https://dljwql.stsci.edu'
    else:
        base_url = 'http://127.0.0.1:8000'

    return base_url


def get_config():
    """Return a dictionary that holds the contents of the ``jwql``
    config file.

    Returns
    -------
    settings : dict
        A dictionary that holds the contents of the config file.
    """
    config_file_location = os.path.join(__location__, 'config.json')

    if not os.path.isfile(config_file_location):
        raise FileNotFoundError('The JWQL package requires a configuration file (config.json) '
                                'to be placed within the jwql/utils directory. '
                                'This file is missing. Please read the relevant wiki page '
                                '(https://github.com/spacetelescope/jwql/wiki/'
                                'Config-file) for more information.')

    with open(config_file_location, 'r') as config_file:
        settings = json.load(config_file)

    return settings


def initialize_instrument_monitor(module):
    """Configures a log file for the instrument monitor run and
    captures the start time of the monitor

    Parameters
    ----------
    module : str
        The module name (e.g. ``dark_monitor``)

    Returns
    -------
    start_time : datetime object
        The start time of the monitor
    log_file : str
        The path to where the log file is stored
    """

    from jwql.utils.logging_functions import configure_logging

    start_time = datetime.datetime.now()
    log_file = configure_logging(module)

    return start_time, log_file


def update_monitor_table(module, start_time, log_file):
    """Update the ``monitor`` database table with information about
    the instrument monitor run

    Parameters
    ----------
    module : str
        The module name (e.g. ``dark_monitor``)
    start_time : datetime object
        The start time of the monitor
    log_file : str
        The path to where the log file is stored
    """

    from jwql.database.database_interface import Monitor

    new_entry = {}
    new_entry['monitor_name'] = module
    new_entry['start_time'] = start_time
    new_entry['end_time'] = datetime.datetime.now()
    new_entry['log_file'] = os.path.basename(log_file)

    Monitor.__table__.insert().execute(new_entry)



#! /usr/bin/env python
"""Prepares plots for Temperature tab

    Module prepares plots for mnemonics below. Combines plots in a grid and
    returns tab object.

    Plot 1 - ASIC 1 Voltages
    IGDP_NRSD_ALG_A1_VDDA
    IGDP_NRSD_ALG_A1GND4VDA
    IGDP_NRSD_ALG_A1GND5VRF
    INRSD_ALG_A1_VDD3P3
    INRSD_ALG_A1_VDD
    INRSD_ALG_A1_REF
    INRSD_A1_DSUB_V
    INRSD_A1_VRESET_V
    INRSD_A1_CELLDRN_V
    INRSD_A1_DRAIN_V
    INRSD_A1_VBIASGATE_V
    INRSD_A1_VBIASPWR_V

    Plot 2 - ASIC 1 Currents
    IGDP_NRSD_ALG_A1_VDD_C
    IGDP_NRSD_ALG_A1VDAP12C
    IGDP_NRSD_ALG_A1VDAN12C
    INRSD_A1_VDDA_I

    Plot 3 - ASIC 2 Voltages
    IGDP_NRSD_ALG_A2_VDDA
    IGDP_NRSD_ALG_A2GND4VDA
    IGDP_NRSD_ALG_A2GND5VRF
    INRSD_ALG_A2_VDD3P3
    INRSD_ALG_A2_VDD
    INRSD_ALG_A2_REF
    INRSD_A2_DSUB_V
    INRSD_A2_VRESET_V
    INRSD_A2_CELLDRN_V
    INRSD_A2_DRAIN_V
    INRSD_A2_VBIASGATE_V
    INRSD_A2_VBIASPWR_V

    Plot 4 - ASIC 2 Currents
    IGDP_NRSD_ALG_A2_VDD_C
    IGDP_NRSD_ALG_A2VDAP12C
    IGDP_NRSD_ALG_A2VDAN12C
    INRSD_A2_VDDA_I

Authors
-------
    - Daniel Kühbacher

Use
---
    The functions within this module are intended to be imported and
    used by ``nirspec_dashboard.py``, e.g.:

    ::
        from .plots.fpa_fpe_tab import fpa_fpe_plots
        tab = fpa_fpe_plots(conn, start, end)

Dependencies
------------
    User must provide database "nirspec_database.db"

"""
import jwql.instrument_monitors.nirspec_monitors.data_trending.utils.sql_interface as sql
import jwql.instrument_monitors.nirspec_monitors.data_trending.plots.plot_functions as pf
from bokeh.models import LinearAxis, Range1d
from bokeh.plotting import figure
from bokeh.models.widgets import Panel, Tabs, Div
from bokeh.models import ColumnDataSource, HoverTool
from bokeh.layouts import WidgetBox, gridplot, Column, Row

import pandas as pd
import numpy as np

from astropy.time import Time


def asic_1_voltages(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",
                toolbar_location = "above",
                plot_width = 1120,
                plot_height = 800,
                x_axis_type = 'datetime',
                output_backend = "webgl",
                x_axis_label = 'Date', y_axis_label='Voltage (V)')

    p.grid.visible = True
    p.title.text = "ASIC 1 Voltages"
    pf.add_basic_layout(p)
    a = pf.add_to_plot(p, "VDDA", "IGDP_NRSD_ALG_A1_VDDA", start, end, conn, color = "burlywood")
    b = pf.add_to_plot(p, "A1GND4VDA", "IGDP_NRSD_ALG_A1GND4VDA", start, end, conn, color = "cadetblue")
    c = pf.add_to_plot(p, "A1GND5VRF", "IGDP_NRSD_ALG_A1GND5VRF", start, end, conn, color = "chartreuse")
    d = pf.add_to_plot(p, "A1VDD3P3", "INRSD_ALG_A1_VDD3P3", start, end, conn, color = "chocolate")
    e = pf.add_to_plot(p, "VDD", "INRSD_ALG_A1_VDD", start, end, conn, color = "coral")
    f = pf.add_to_plot(p, "REF", "INRSD_ALG_A1_REF", start, end, conn, color = "darkorange")
    g = pf.add_to_plot(p, "DSUB_V", "INRSD_A1_DSUB_V", start, end, conn, color = "crimson")
    h = pf.add_to_plot(p, "VRESET_V", "INRSD_A1_VRESET_V", start, end, conn, color = "cyan")
    i = pf.add_to_plot(p, "CELLDRN_V", "INRSD_A1_CELLDRN_V", start, end, conn, color = "darkblue")
    j = pf.add_to_plot(p, "DRAIN_V", "INRSD_A1_DRAIN_V", start, end, conn, color = "darkgreen")
    k = pf.add_to_plot(p, "VBIASGATE_V", "INRSD_A1_VBIASGATE_V", start, end, conn, color = "darkmagenta")
    l = pf.add_to_plot(p, "VBIASPWR_V", "INRSD_A1_VBIASPWR_V", start, end, conn, color = "cornflowerblue")
    pf.add_hover_tool(p,[a,b,c,d,e,f,g,h,i,j,k,l])
    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"

    return p

def asic_2_voltages(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",
                toolbar_location = "above",
                plot_width = 1120,
                plot_height = 800,
                x_axis_type = 'datetime',
                output_backend = "webgl",
                x_axis_label = 'Date', y_axis_label='Voltage (V)')

    p.grid.visible = True
    p.title.text = "ASIC 2 Voltages"
    pf.add_basic_layout(p)
    a = pf.add_to_plot(p, "VDDA", "IGDP_NRSD_ALG_A2_VDDA", start, end, conn, color = "burlywood")
    b = pf.add_to_plot(p, "A2GND4VDA", "IGDP_NRSD_ALG_A2GND4VDA", start, end, conn, color = "cadetblue")
    c = pf.add_to_plot(p, "A2GND5VRF", "IGDP_NRSD_ALG_A2GND5VRF", start, end, conn, color = "chartreuse")
    d = pf.add_to_plot(p, "A2VDD3P3", "INRSD_ALG_A2_VDD3P3", start, end, conn, color = "chocolate")
    e = pf.add_to_plot(p, "VDD", "INRSD_ALG_A2_VDD", start, end, conn, color = "coral")
    f = pf.add_to_plot(p, "REF", "INRSD_ALG_A2_REF", start, end, conn, color = "darkorange")
    g = pf.add_to_plot(p, "DSUB_V", "INRSD_A2_DSUB_V", start, end, conn, color = "crimson")
    h = pf.add_to_plot(p, "VRESET_V", "INRSD_A2_VRESET_V", start, end, conn, color = "cyan")
    i = pf.add_to_plot(p, "CELLDRN_V", "INRSD_A2_CELLDRN_V", start, end, conn, color = "darkblue")
    j = pf.add_to_plot(p, "DRAIN_V", "INRSD_A2_DRAIN_V", start, end, conn, color = "darkgreen")
    k = pf.add_to_plot(p, "VBIASGATE_V", "INRSD_A2_VBIASGATE_V", start, end, conn, color = "darkmagenta")
    l = pf.add_to_plot(p, "VBIASPWR_V", "INRSD_A2_VBIASPWR_V", start, end, conn, color = "cornflowerblue")
    pf.add_hover_tool(p,[a,b,c,d,e,f,g,h,i,j,k,l])
    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"

    return p

def asic_1_currents(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",
                toolbar_location = "above",
                plot_width = 560,
                plot_height = 500,
                x_axis_type = 'datetime',
                output_backend = "webgl",
                x_axis_label = 'Date', y_axis_label='Current (mA)')

    p.grid.visible = True
    p.title.text = "ASIC 1 Currents"
    pf.add_basic_layout(p)

    a = pf.add_to_plot(p, "VDD_C", "IGDP_NRSD_ALG_A1_VDD_C", start, end, conn, color = "burlywood")
    b = pf.add_to_plot(p, "A1VDAP12C", "IGDP_NRSD_ALG_A1VDAP12C", start, end, conn, color = "cadetblue")
    c = pf.add_to_plot(p, "A1VDAN12C", "IGDP_NRSD_ALG_A1VDAN12C", start, end, conn, color = "chartreuse")
    d = pf.add_to_plot(p, "VDDA_I", "INRSD_A1_VDDA_I", start, end, conn, color = "chocolate")

    pf.add_hover_tool(p,[a,b,c,d])

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"
    p.legend.orientation = "horizontal"

    return p

def asic_2_currents(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",
                toolbar_location = "above",
                plot_width = 560,
                plot_height = 500,
                x_axis_type = 'datetime',
                output_backend = "webgl",
                x_axis_label = 'Date', y_axis_label='Current (mA)')

    p.grid.visible = True
    p.title.text = "ASIC 2 Currents"
    pf.add_basic_layout(p)

    a = pf.add_to_plot(p, "VDD_C", "IGDP_NRSD_ALG_A2_VDD_C", start, end, conn, color = "burlywood")
    b = pf.add_to_plot(p, "A2VDAP12C", "IGDP_NRSD_ALG_A2VDAP12C", start, end, conn, color = "cadetblue")
    c = pf.add_to_plot(p, "A2VDAN12C", "IGDP_NRSD_ALG_A2VDAN12C", start, end, conn, color = "chartreuse")
    d = pf.add_to_plot(p, "VDDA_I", "INRSD_A2_VDDA_I", start, end, conn, color = "chocolate")

    pf.add_hover_tool(p,[a,b,c,d])

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"
    p.legend.orientation = "horizontal"
    return p


def fpe_fpa_plots(conn, start, end):
    '''Combines plots to a tab
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : tab object
        used by dashboard.py to set up dashboard
    '''
    descr = Div(text=
    """
    <style>
    table, th, td {
      border: 1px solid black;
      background-color: #efefef;
      border-collapse: collapse;
      padding: 5px
    }
    table {
      border-spacing: 15px;
    }
    </style>

    <body>
    <table style="width:100%">
      <tr>
        <th><h6>Plotname</h6></th>
        <th><h6>Mnemonic</h6></th>
        <th><h6>Description</h6></th>
      </tr>
      <tr>
        <td>ASIC (1,2) Voltages</td>
        <td>IGDP_NRSD_ALG_A(1,2)_VDDA<br>
            IGDP_NRSD_ALG_A(1,2)GND4VDA<br>
            IGDP_NRSD_ALG_A(1,2)GND5VRF<br>
            INRSD_ALG_A(1,2)_VDD3P3<br>
            INRSD_ALG_A(1,2)_VDD<br>
            INRSD_ALG_A(1,2)_REF<br>
            INRSD_A(1,2)_DSUB_V<br>
            INRSD_A(1,2)_VRESET_V<br>
            INRSD_A(1,2)_CELLDRN_V<br>
            INRSD_A(1,2)_DRAIN_V<br>
            INRSD_A(1,2)_VBIASGATE_V<br>
            INRSD_A(1,2)_VBIASPWR_V<br>
        </td>
        <td>
            ASIC (1,2) VDDA Voltage<br>
            ASIC (1,2) VDDA/Ground Voltage<br>
            ASIC (1,2) Ref/Ground Voltage<br>
            ASIC (1,2) VDD 3.3 Supply Voltage<br>
            ASIC (1,2) VDD Voltage<br>
            ASIC (1,2) Reference Voltage<br>
            ASIC (1,2) Dsub Voltage<br>
            ASIC (1,2) Reset Voltage<br>
            ASIC (1,2) Cell Drain Voltage<br>
            ASIC (1,2) Drain Voltage<br>
            ASIC (1,2) Bias Gate Voltage<br>
            ASIC (1,2) Bias Power Voltage<br>
        </td>
      </tr>

      <tr>
        <td>ASIC (1,2) Currents</td>
        <td>IGDP_NRSD_ALG_A(1,2)_VDD_C<br>
            IGDP_NRSD_ALG_A(1,2)VDAP12C<br>
            IGDP_NRSD_ALG_A(1,2)VDAN12C<br>
            INRSD_A(1,2)_VDDA_I<br>
        </td>
        <td>ASIC (1,2) VDD Current<br>
            ASIC (1,2) VDDA +12V Current<br>
            ASIC (1,2) VDDA -12V Current<br>
            ASIC (1,2) VDDA Current<br>
        </td>
      </tr>
      
    </table>
    </body>
    """, width=1100)

    plot1 = asic_1_voltages(conn, start, end)
    plot2 = asic_2_voltages(conn, start, end)
    plot3 = asic_1_currents(conn, start, end)
    plot4 = asic_2_currents(conn, start, end)

    currents = Row(plot3, plot4)
    layout = Column(descr, plot1, plot2, currents)

    tab = Panel(child = layout, title = "FPE/FPA")

    return tab



#!/usr/bin/env python

"""Tests for the ``logging_functions`` module.

Authors
-------

    - Matthew Bourque

Use
---

    These tests can be run via the command line (omit the -s to
    suppress verbose output to stdout):

    ::

        pytest -s test_logging_functions.py
"""

import logging
import os
import pytest
import shutil

from jwql.utils import logging_functions
from jwql.utils.logging_functions import configure_logging, log_fail, log_info, make_log_file
from jwql.utils.utils import get_config


@log_fail
@log_info
def perform_basic_logging():
    """Performs some basic logging to the test log file"""

    logging.info('This is some logging info')
    logging.warning('This is a normal warning')
    logging.critical('This is a critical warning')


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_configure_logging():
    """Assert that the ``configure_logging`` function successfully
    creates a log file"""

    log_file = logging_functions.configure_logging('test_logging_functions')
    assert os.path.exists(log_file)

    # Remove the log file
    shutil.rmtree(os.path.dirname(log_file), ignore_errors=True)


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_make_log_file():
    """Assert that ``make_log_file`` function returns the appropriate
    path for a log file"""

    module = 'test_logging_functions'
    log_file = make_log_file(module)

    correct_locations = [
        os.path.join(get_config()['log_dir'], 'dev', module, os.path.basename(log_file)),
        os.path.join(get_config()['log_dir'], 'test', module, os.path.basename(log_file)),
        os.path.join(get_config()['log_dir'], 'prod', module, os.path.basename(log_file))]

    assert log_file in correct_locations


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_logging_functions():
    """A generic end-to-end test that creates a log file, does some
    basic logging, then asserts that some logging content exists"""

    log_file = configure_logging('test_logging_functions')
    perform_basic_logging()

    # Open the log file and make some assertions
    with open(log_file, 'r') as f:
        data = f.readlines()
    data = str([line.strip() for line in data])
    testable_content = ['User:', 'System:', 'Python Executable Path:', 'INFO:', 'WARNING:', 'CRITICAL:',
                        'Elapsed Real Time:', 'Elapsed CPU Time:', 'Completed Successfully']
    for item in testable_content:
        assert item in data



#! /usr/bin/env python

"""Tests for the ``instrument_properties`` module.

Authors
-------

    - Bryan Hilbert

Use
---

    These tests can be run via the command line (omit the ``-s`` to
    suppress verbose output to stdout):
    ::

        pytest -s test_instrument_properties.py
"""

import os
import pytest

import numpy as np

from jwql.utils import instrument_properties
from jwql.utils.utils import get_config


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_amplifier_info():
    """Test that the correct number of amplifiers are found for a given
    file
    """

    data_dir = os.path.join(get_config()['test_dir'], 'dark_monitor')

    fullframe = instrument_properties.amplifier_info(os.path.join(data_dir, 'test_image_ff.fits'))
    fullframe_truth = (4, {'1': [(4, 4), (512, 2044)],
                           '2': [(512, 4), (1024, 2044)],
                           '3': [(1024, 4), (1536, 2044)],
                           '4': [(1536, 4), (2044, 2044)]})
    assert fullframe == fullframe_truth

    fullframe = instrument_properties.amplifier_info(os.path.join(data_dir, 'test_image_ff.fits'), omit_reference_pixels=False)
    fullframe_truth = (4, {'1': [(0, 0), (512, 2048)],
                           '2': [(512, 0), (1024, 2048)],
                           '3': [(1024, 0), (1536, 2048)],
                           '4': [(1536, 0), (2048, 2048)]})
    assert fullframe == fullframe_truth

    subarray = instrument_properties.amplifier_info(os.path.join(data_dir, 'test_image_1.fits'))
    subarray_truth = (1, {'1': [(0, 0), (10, 10)]})
    assert subarray == subarray_truth

    subarray_one = instrument_properties.amplifier_info(os.path.join(data_dir, 'test_image_grismstripe_one_amp.fits'))
    subarray_one_truth = (1, {'1': [(4, 4), (2044, 64)]})
    assert subarray_one == subarray_one_truth

    subarray_four = instrument_properties.amplifier_info(os.path.join(data_dir, 'test_image_grismstripe_four_amp.fits'))
    subarray_four_truth = (4, {'1': [(4, 4), (512, 64)],
                               '2': [(512, 4), (1024, 64)],
                               '3': [(1024, 4), (1536, 64)],
                               '4': [(1536, 4), (2044, 64)]})
    assert subarray_four == subarray_four_truth


def test_calc_frame_time():
    """Test calcuation of frametime for a given instrument/aperture"""

    nearir_fullframe = 10.73677
    nircam_160 = 0.27864
    nrc_fullframe = instrument_properties.calc_frame_time('nircam', 'NRCA1_FULL', 2048, 2048, 4)
    nrc_160 = instrument_properties.calc_frame_time('nircam', 'NRCA1_SUB160', 160, 160, 1)
    nrs_fullframe = instrument_properties.calc_frame_time('niriss', 'NIS_CEN', 2048, 2048, 4)
    #nrs_some_subarra = instrument_properies.calc_frame_time('niriss', '????', ??, ??, ?)

    print('STILL NEED TO ADD FRAMETIME CALCS FOR MIRI AND NIRSPEC TO THE CALC_FRAME_TIME_FUNCTION')
    print('CONFIRM NIRCAMSUB160 TIME ON JDOX')

    assert np.isclose(nrc_fullframe, nearir_fullframe, atol=0.001, rtol=0)
    assert np.isclose(nrc_160, nircam_160, atol=0.001, rtol=0)
    assert np.isclose(nrs_fullframe, nearir_fullframe, atol=0.001, rtol=0)



"""Provides functions that define context inherent to all views.

The functions within this module define ``context`` that will be
included in requests, in addition to any specific ``context`` provided
in the view.

Authors
-------

    - Matthew Bourque

Use
---

    This module is defined under the ``TEMPLATES.OPTIONS`` setting in
    ``settings.py``, e.g.:
    ::

        TEMPLATES = [
            {'OPTIONS': {'context_processors': ['jwql.website.apps.jwql.context_processors.base_context'],},}
        ]

    As such, it will automatically be executed upon each request.
"""

import jwql
from jwql.utils.constants import JWST_INSTRUMENT_NAMES, MONITORS

from .oauth import auth_info


@auth_info
def base_context(request, user):
    """Provide the context needed for the ``base.html`` template.

    Parameters
    ----------
    request : HttpRequest object
        Incoming request from the webpage

    Returns
    -------
    context : dict
        A dictionary containing data needed to render the ``base.html``
        template
    """

    context = {}
    context['inst_list'] = JWST_INSTRUMENT_NAMES
    context['tools'] = MONITORS
    context['user'] = user
    context['version'] = jwql.__version__

    return context



#! /usr/bin/env python

"""
This module is a collection of plotting functions that may be used
across the ``jwql`` application.

Authors:
--------

    - Joe Filippazzo

Use:
----

    This module can be use as follows:

    ::

        from jwql.utils import plotting
        from pandas import DataFrame
        data = DataFrame({'meow': {'foo': 12, 'bar': 23, 'baz': 2},
                          'mix': {'foo': 45, 'bar': 31, 'baz': 23},
                          'deliver': {'foo': 62, 'bar': 20, 'baz': 9}})
        data = data.reset_index()
        plt = plotting.bar_chart(data, 'index')
"""

from bokeh.models import ColumnDataSource, FactorRange, HoverTool
from bokeh.palettes import Category20c
from bokeh.plotting import figure
from bokeh.transform import factor_cmap


def bar_chart(dataframe, groupcol, datacols=None, **kwargs):
    """Create a pie chart from a Pandas DataFrame

    Parameters
    ----------
    dataframe : pandas.DataFrame
        A dataframe of values
    groupcol : str
        The name of the column with the group labels
    datacol : str, sequence (optional)
        The name or list of names of the column containing the data.
        In None, uses all columns except **groupcol**

    Returns
    -------
    plt : obj
        The generated bokeh.figure object
    """
    
    # Get the groups
    groups = list(dataframe[groupcol])

    # Get the datacols
    if datacols is None:
        datacols = [col for col in list(dataframe.columns) if col != groupcol]

    # Make a dictionary of the groups and data
    data = {'groups': groups}
    for col in datacols:
        data.update({col: list(dataframe[col])})

    # hstack it
    x = [(group, datacol) for group in groups for datacol in datacols]
    counts = sum(zip(*[data[col] for col in datacols]), ())
    colors = max(3, len(datacols))
    source = ColumnDataSource(data=dict(x=x, counts=counts))

    # Make the figure
    hover = HoverTool(tooltips=[('count', '@counts')])
    plt = figure(x_range=FactorRange(*x), plot_height=250, tools=[hover],
                 **kwargs)
    plt.vbar(x='x', top='counts', width=0.9, source=source, line_color="white",
             fill_color=factor_cmap('x', palette=Category20c[colors],
                                    factors=datacols, start=1, end=2))

    # Formatting
    plt.y_range.start = 0
    plt.x_range.range_padding = 0.1
    plt.xaxis.major_label_orientation = 1
    plt.xgrid.grid_line_color = None

    return plt



#! /usr/bin/env python
"""Prepares plots for POWER tab

    Module prepares plots for mnemonics below. Combines plots in a grid and
    returns tab object.

    Plot 1 - ICE Power Data
    GP_ZPSVOLT
    SE_ZINRSICEA / SE_ZINRSICEB
    INRSH_HK_P15V
    INRSH_HK_N15V
    INRSH_HK_VMOTOR
    INRSH_HK_P5V
    INRSH_HK_2P5V
    INRSH_HK_ADCTGAIN
    INRSH_HK_ADCTOFFSET
    INRSH_OA_VREFOFF
    INRSH_OA_VREF

    Plot 2 - MCE Power Data
    GP_ZPSVOLT
    SE_ZINRSMCEA / SE_ZINRSMCEB

    Plot 3 - FPE Power Data
    GP_ZPSVOLT
    SE_ZINRSFPEA / SE_ZINRSFPEB
    INRSD_ALG_ACC_P12C
    INRSD_ALG_ACC_N12C
    INRSD_ALG_ACC_3D3_1D5_C
    INRSD_ALG_CHASSIS

Authors
-------
    - Daniel Kühbacher

Use
---
    The functions within this module are intended to be imported and
    used by ``nirspec_dashboard.py``, e.g.:

    ::
        from .plots.power_tab import power_plots
        tab = power_plots(conn, start, end)

Dependencies
------------
    User must provide database "miri_database.db"

"""
import jwql.instrument_monitors.nirspec_monitors.data_trending.utils.sql_interface as sql
import jwql.instrument_monitors.nirspec_monitors.data_trending.plots.plot_functions as pf
from bokeh.models import LinearAxis, Range1d
from bokeh.plotting import figure
from bokeh.models.widgets import Panel, Tabs, Div
from bokeh.models import ColumnDataSource, HoverTool
from bokeh.layouts import WidgetBox, gridplot, Column

import pandas as pd
import numpy as np

from astropy.time import Time


def ice_power(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",
                toolbar_location = "above",
                plot_width = 1120,
                plot_height = 700,
                x_axis_type = 'datetime',
                y_range = [-20, 20],
                output_backend = "webgl",
                x_axis_label = 'Date', y_axis_label='Voltage (V)')

    p.grid.visible = True
    p.title.text = "ICE Power Parameters"
    pf.add_basic_layout(p)

    p.extra_y_ranges = {"current": Range1d(start = 0, end=0.8)}
    #a = pf.add_to_plot(p, "In_VOlt", "GP_ZPSVOLT", start, end, conn, color = "red")
    b = pf.add_to_plot(p, "ICE A current", "SE_ZINRSICEA", start, end, conn, color = "blue", y_axis="current")
    c = pf.add_to_plot(p, "P15V", "INRSH_HK_P15V", start, end, conn, color = "red")
    d = pf.add_to_plot(p, "N15V", "INRSH_HK_N15V", start, end, conn, color = "orange")
    e = pf.add_to_plot(p, "VMOTOR", "INRSH_HK_VMOTOR", start, end, conn, color = "burlywood")
    f = pf.add_to_plot(p, "P5V", "INRSH_HK_P5V", start, end, conn, color = "green")
    g = pf.add_to_plot(p, "2P5V", "INRSH_HK_2P5V", start, end, conn, color = "darkgreen")
    h = pf.add_to_plot(p, "ADCTGAIN", "INRSH_HK_ADCTGAIN", start, end, conn, color = "brown")
    i = pf.add_to_plot(p, "ADCOFFSET", "INRSH_HK_ADCTOFFSET", start, end, conn, color = "navy")
    p.add_layout(LinearAxis(y_range_name = "current", axis_label = "Current (A)", axis_label_text_color = "blue"), 'right')

    pf.add_hover_tool(p,[b,c,d,e,g,f,h,i])

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"
    p.legend.orientation = "horizontal"

    return p

def mce_power(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",
                toolbar_location = "above",
                plot_width = 1120,
                plot_height = 400,
                x_axis_type = 'datetime',
                output_backend = "webgl",
                x_axis_label = 'Date', y_axis_label='Current (A)')

    p.grid.visible = True
    p.title.text = "MCE Power Parameters"
    pf.add_basic_layout(p)

    b = pf.add_to_plot(p, "MCE A current", "SE_ZINRSMCEA", start, end, conn, color = "blue")

    pf.add_hover_tool(p,[b])

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"

    return p

def fpe_power(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",
                toolbar_location = "above",
                plot_width = 1120,
                plot_height = 700,
                y_range = [-30,280],
                x_axis_type = 'datetime',
                output_backend = "webgl",
                x_axis_label = 'Date', y_axis_label='Voltage (V)')

    p.grid.visible = True
    p.title.text = "FPE Power Parameters"
    pf.add_basic_layout(p)


    p.extra_y_ranges = {"current": Range1d(start = 0, end=0.8)}
    #a = pf.add_to_plot(p, "In_VOlt", "GP_ZPSVOLT", start, end, conn, color = "red")
    b = pf.add_to_plot(p, "FPE A current", "SE_ZINRSFPEA", start, end, conn, color = "blue", y_axis="current")
    c = pf.add_to_plot(p, "P12C", "INRSD_ALG_ACC_P12C", start, end, conn, color = "red")
    d = pf.add_to_plot(p, "N15V", "INRSH_HK_N15V", start, end, conn, color = "orange")
    e = pf.add_to_plot(p, "N12C", "INRSD_ALG_ACC_N12C", start, end, conn, color = "burlywood")
    f = pf.add_to_plot(p, "1D5", "INRSD_ALG_ACC_3D3_1D5_C", start, end, conn, color = "green")
    g = pf.add_to_plot(p, "Chassis", "INRSD_ALG_CHASSIS", start, end, conn, color = "purple")
    p.add_layout(LinearAxis(y_range_name = "current", axis_label = "Current (A)", axis_label_text_color = "blue"), 'right')

    pf.add_hover_tool(p,[b,c,d,e,f,g])

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"

    return p



def power_plots(conn, start, end):
    '''Combines plots to a tab
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : tab object
        used by dashboard.py to set up dashboard
    '''
    descr = Div(text=
    """
    <style>
    table, th, td {
      border: 1px solid black;
      background-color: #efefef;
      border-collapse: collapse;
      padding: 5px
    }
    table {
      border-spacing: 15px;
    }
    </style>

    <body>
    <table style="width:100%">
      <tr>
        <th><h6>Plotname</h6></th>
        <th><h6>Mnemonic</h6></th>
        <th><h6>Description</h6></th>
      </tr>
      <tr>
        <td>ICE Power Parameters</td>
        <td>GP_ZPSVOLT (missing)<br>
            SE_ZINRSICEA<br>
            INRSH_HK_P15V<br>
            INRSH_HK_N15V<br>
            INRSH_HK_VMOTOR<br>
            INRSH_HK_P5V<br>
            INRSH_HK_2P5V<br>
            INRSH_HK_ADCTGAIN<br>
            INRSH_HK_ADCTOFFSET<br>
            INRSH_OA_VREFOFF<br>
            INRSH_OA_VREF<br>
        </td>
        <td>ICE Input Voltage<br>
            ICE Input Current (A side)<br>
            ICE +15V Voltage<br>
            ICE -15V Voltage<br>
            ICE Motor Voltage<br>
            ICE +5V FPGA Voltage<br>
            ICE +2V5 FPGA Voltage<br>
            ICE ADC TM Chain Gain for Calibration<br>
            ICE ADC TM Chain Offset for Calibration<br>
        </td>
      </tr>

      <tr>
        <td>MCE Power Parameters</td>
        <td>GP_ZPSVOLT (missing)<br>
            SE_ZINRSMCEA
        </td>
        <td>ICE Input Voltage<br>
            MCE Input Current (A side)<br>
        </td>
      </tr>

      <tr>
        <td>FPE Power Parameters</td>
        <td>GP_ZPSVOLT (missing)<br>
            SE_ZINRSFPEA<br>
            INRSD_ALG_ACC_P12C<br>
            INRSD_ALG_ACC_N12C<br>
            INRSD_ALG_ACC_3D3_1D5_C<br>
            INRSD_ALG_CHASSIS<br>
        </td>
        <td>ICE Input Voltage<br>
            MCE Input Current (A side)<br>
            ACC +12V Current<br>
            ACC -12V Current<br>
            ACC 3.3/1.5 Supply Current<br>
            Chassis Voltage<br>
        </td>
      </tr>

    </table>
    </body>
    """, width=1100)

    plot1 = ice_power(conn, start, end)
    plot2 = mce_power(conn, start, end)
    plot3 = fpe_power(conn, start, end)

    layout = Column(descr, plot1, plot2, plot3)

    tab = Panel(child = layout, title = "POWER")

    return tab



#! /usr/bin/env python
"""Prepares plots for FPE VOLTAGE tab

    Module prepares plots for mnemonics below. Combines plots in a grid and
    returns tab object.

    Plot 1:
    IMIR_HK_FW_POS_RATIO_FND
    IMIR_HK_FW_POS_RATIO_OPAQUE
    IMIR_HK_FW_POS_RATIO_F1000W
    IMIR_HK_FW_POS_RATIO_F1130W
    IMIR_HK_FW_POS_RATIO_F1280W
    IMIR_HK_FW_POS_RATIO_P750L
    IMIR_HK_FW_POS_RATIO_F1500W
    IMIR_HK_FW_POS_RATIO_F1800W
    IMIR_HK_FW_POS_RATIO_F2100W
    IMIR_HK_FW_POS_RATIO_F560W
    IMIR_HK_FW_POS_RATIO_FLENS
    IMIR_HK_FW_POS_RATIO_F2300C
    IMIR_HK_FW_POS_RATIO_F770W
    IMIR_HK_FW_POS_RATIO_F1550C
    IMIR_HK_FW_POS_RATIO_F2550W
    IMIR_HK_FW_POS_RATIO_F1140C
    IMIR_HK_FW_POS_RATIO_F2550WR
    IMIR_HK_FW_POS_RATIO_F1065C

    Plot 2:
    IMIR_HK_GW14_POS_RATIO_SHORT
    IMIR_HK_GW14_POS_RATIO_MEDIUM
    IMIR_HK_GW14_POS_RATIO_LONG

    Plot 3:
    IMIR_HK_GW23_POS_RATIO_SHORT
    IMIR_HK_GW23_POS_RATIO_MEDIUM
    IMIR_HK_GW23_POS_RATIO_LONG

    Plot 4:
    IMIR_HK_CCC_POS_RATIO_LOCKED
    IMIR_HK_CCC_POS_RATIO_OPEN
    IMIR_HK_CCC_POS_RATIO_CLOSED

Authors
-------
    - Daniel Kühbacher

Use
---
    The functions within this module are intended to be imported and
    used by ``dashboard.py``, e.g.:

    ::
        from .plots.wheel_ratio_tab import wheel_plots
        tab = wheel_plots(conn, start, end)

Dependencies
------------
    User must provide database "miri_database.db"

"""

import jwql.instrument_monitors.miri_monitors.data_trending.utils.sql_interface as sql
import jwql.instrument_monitors.miri_monitors.data_trending.plots.plot_functions as pf
import jwql.instrument_monitors.miri_monitors.data_trending.utils.mnemonics as mn
from bokeh.plotting import figure
from bokeh.models import BoxAnnotation, LinearAxis, Range1d
from bokeh.embed import components
from bokeh.models.widgets import Panel, Tabs, Div
from bokeh.models import ColumnDataSource
from bokeh.layouts import column, row, WidgetBox

import pandas as pd
import numpy as np

from astropy.time import Time


def gw14(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",       \
                toolbar_location = "above",                         \
                plot_width = 1120,                                   \
                plot_height = 500,                                  \
                y_range = [-2,2],                                   \
                x_axis_type = 'datetime',                           \
                x_axis_label = 'Date', y_axis_label = 'ratio (normalized)')

    p.grid.visible = True
    p.title.text = "DGA-A Ratio"
    p.title.align = "left"
    pf.add_basic_layout(p)

    pf.add_to_wplot(p, "SHORT", "IMIR_HK_GW14_POS_RATIO_SHORT", start, end, conn, mn.gw14_nominals['SHORT'], color = "green")
    pf.add_to_wplot(p, "MEDIUM", "IMIR_HK_GW14_POS_RATIO_MEDIUM", start, end, conn, mn.gw14_nominals['MEDIUM'], color = "red")
    pf.add_to_wplot(p, "LONG", "IMIR_HK_GW14_POS_RATIO_LONG", start, end, conn, mn.gw14_nominals['LONG'], color = "blue")

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"

    return p


def gw23(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",       \
                toolbar_location = "above",                         \
                plot_width = 1120,                                   \
                plot_height = 500,                                  \
                y_range = [-2,2],                                   \
                x_axis_type = 'datetime',                           \
                x_axis_label = 'Date', y_axis_label = 'ratio (normalized)')

    p.grid.visible = True
    p.title.text = "DGA-B Ratio"
    p.title.align = "left"
    pf.add_basic_layout(p)

    pf.add_to_wplot(p, "SHORT", "IMIR_HK_GW23_POS_RATIO_SHORT", start, end, conn, mn.gw23_nominals['SHORT'], color = "green")
    pf.add_to_wplot(p, "MEDIUM", "IMIR_HK_GW23_POS_RATIO_MEDIUM", start, end, conn, mn.gw23_nominals['MEDIUM'], color = "red")
    pf.add_to_wplot(p, "LONG", "IMIR_HK_GW23_POS_RATIO_LONG", start, end, conn, mn.gw23_nominals['LONG'], color = "blue")

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"

    return p


def ccc(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",       \
                toolbar_location = "above",                         \
                plot_width = 1120,                                   \
                plot_height = 500,                                  \
                y_range = [-2,2],                                   \
                x_axis_type = 'datetime',                           \
                x_axis_label = 'Date', y_axis_label = 'ratio (normalized)')

    p.grid.visible = True
    p.title.text = "CCC Ratio"
    pf.add_basic_layout(p)

    #add_to_wplot(p, "LOCKED", "IMIR_HK_CCC_POS_RATIO_LOCKED", start, end, conn, mn.ccc_nominals['LOCKED'], color = "green")
    pf.add_to_wplot(p, "OPEN", "IMIR_HK_CCC_POS_RATIO_OPEN", start, end, conn, mn.ccc_nominals['OPEN'], color = "red")
    pf.add_to_wplot(p, "CLOSED", "IMIR_HK_CCC_POS_RATIO_CLOSED", start, end, conn, mn.ccc_nominals['CLOSED'], color = "blue")

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"

    return p


def fw(conn, start, end):
    '''Create specific plot and return plot object
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : Plot object
        Bokeh plot
    '''

    # create a new plot with a title and axis labels
    p = figure( tools = "pan,wheel_zoom,box_zoom,reset,save",       \
                toolbar_location = "above",                         \
                plot_width = 1120,                                   \
                plot_height = 800,                                  \
                y_range = [-6,4],                                   \
                x_axis_type = 'datetime',                           \
                x_axis_label = 'Date', y_axis_label='ratio (normalized)')

    p.grid.visible = True
    p.title.text = "Filterwheel Ratio"
    pf.add_basic_layout(p)

    pf.add_to_wplot(p, "FND", "IMIR_HK_FW_POS_RATIO_FND", start, end, conn, mn.fw_nominals['FND'], color = "green")
    pf.add_to_wplot(p, "OPAQUE", "IMIR_HK_FW_POS_RATIO_OPAQUE", start, end, conn, mn.fw_nominals['OPAQUE'], color = "red")
    pf.add_to_wplot(p, "F1000W", "IMIR_HK_FW_POS_RATIO_F1000W", start, end, conn, mn.fw_nominals['F1000W'], color = "blue")
    pf.add_to_wplot(p, "F1130W", "IMIR_HK_FW_POS_RATIO_F1130W", start, end, conn, mn.fw_nominals['F1130W'], color = "orange")
    pf.add_to_wplot(p, "F1280W", "IMIR_HK_FW_POS_RATIO_F1280W", start, end, conn, mn.fw_nominals['F1280W'], color = "firebrick")
    pf.add_to_wplot(p, "P750L", "IMIR_HK_FW_POS_RATIO_P750L", start, end, conn, mn.fw_nominals['P750L'], color = "cyan")
    pf.add_to_wplot(p, "F1500W", "IMIR_HK_FW_POS_RATIO_F1500W", start, end, conn, mn.fw_nominals['F1500W'], color = "magenta")
    pf.add_to_wplot(p, "F1800W", "IMIR_HK_FW_POS_RATIO_F1800W", start, end, conn, mn.fw_nominals['F1800W'], color = "burlywood")
    pf.add_to_wplot(p, "F2100W", "IMIR_HK_FW_POS_RATIO_F2100W", start, end, conn, mn.fw_nominals['F2100W'], color = "cadetblue")
    pf.add_to_wplot(p, "F560W", "IMIR_HK_FW_POS_RATIO_F560W", start, end, conn, mn.fw_nominals['F560W'], color = "chartreuse")
    pf.add_to_wplot(p, "FLENS", "IMIR_HK_FW_POS_RATIO_FLENS", start, end, conn, mn.fw_nominals['FLENS'], color = "brown")
    pf.add_to_wplot(p, "F2300C", "IMIR_HK_FW_POS_RATIO_F2300C", start, end, conn, mn.fw_nominals['F2300C'], color = "chocolate")
    pf.add_to_wplot(p, "F770W", "IMIR_HK_FW_POS_RATIO_F770W", start, end, conn, mn.fw_nominals['F770W'], color = "darkorange")
    pf.add_to_wplot(p, "F1550C", "IMIR_HK_FW_POS_RATIO_F1550C", start, end, conn, mn.fw_nominals['F1550C'], color = "darkgreen")
    pf.add_to_wplot(p, "F2550W", "IMIR_HK_FW_POS_RATIO_F2550W", start, end, conn, mn.fw_nominals['F2550W'], color = "darkcyan")
    pf.add_to_wplot(p, "F1140C", "IMIR_HK_FW_POS_RATIO_F1140C", start, end, conn, mn.fw_nominals['F1140C'], color = "darkmagenta")
    pf.add_to_wplot(p, "F2550WR", "IMIR_HK_FW_POS_RATIO_F2550WR", start, end, conn, mn.fw_nominals['F2550WR'], color = "crimson")
    pf.add_to_wplot(p, "F1065C", "IMIR_HK_FW_POS_RATIO_F1065C", start, end, conn, mn.fw_nominals['F1065C'], color = "cornflowerblue")

    p.legend.location = "bottom_right"
    p.legend.click_policy = "hide"

    return p


def wheel_ratios(conn, start, end):
    '''Combine plots to a tab
    Parameters
    ----------
    conn : DBobject
        Connection object that represents database
    start : time
        Startlimit for x-axis and query (typ. datetime.now()- 4Months)
    end : time
        Endlimit for x-axis and query (typ. datetime.now())
    Return
    ------
    p : tab object
        used by dashboard.py to set up dashboard
    '''

    descr = Div(text=
    """
    <style>
    table, th, td {
      border: 1px solid black;
      background-color: #efefef;
      border-collapse: collapse;
      padding: 5px
    }
    </style>

    <body>
    <table style="width:100%">
      <tr>
        <th><h6>Plotname</h6></th>
        <th><h6>Mnemonic</h6></th>
        <th><h6>Description</h6></th>
      </tr>
      <tr>
        <td>Filterwheel Ratio</td>
        <td>IMIR_HK_FW_POS_RATIO<br>
            IMIR_HK_FW_CUR_POS<br></td>
        <td>FW position sensor ratio (normalised) and commanded position</td>
      </tr>
      <tr>
        <td>DGA-A Ratio</td>
        <td>IMIR_HK_GW14_POS_RATIO<br>
            IMIR_HK_GW14_CUR_POS<br></td>
        <td>DGA-A position sensor ratio (normalised) and commanded position</td>
      </tr>
      <tr>
        <td>DGA-B Ratio</td>
        <td>IMIR_HK_GW23_POS_RATIO<br>
            IMIR_HK_GW23_CUR_POS<br></td>
        <td>DGA-B position sensor ratio (normalised) and commanded position</td>
      </tr>
      <tr>
        <td>CCC Ratio</td>
        <td>IMIR_HK_CCC_POS_RATIO<br>
            IMIR_HK_CCC_CUR_POS<br></td>
        <td>Contamination Control Cover position sensor ratio (normalised) and commanded position</td>
      </tr>
    </table>
    </body>
    """, width=1100)

    plot1 = fw(conn, start, end)
    plot2 = gw14(conn, start, end)
    plot3 = gw23(conn,  start, end)
    plot4 = ccc(conn, start, end)

    layout = column(descr, plot1, plot2, plot3, plot4)
    tab = Panel(child = layout, title = "WHEEL RATIO")

    return tab



"""Customizes the ``jwql`` web app administrative page.

** CURRENTLY NOT IN USE **

Used to customize django's admin interface, and how the data contained
in specific models is portrayed.

Authors
-------

    - Lauren Chambers

References
----------

    For more information please see:
        ``https://docs.djangoproject.com/en/2.0/ref/contrib/admin/``
"""

from django.contrib import admin

from .models import ImageData


class ImageDataAdmin(admin.ModelAdmin):
    # fieldsets = [('Filepath', {'fields': ['filepath']}),
    # 			 ('Instrument', {'fields': ['inst']}),
    #              ('Date information', {'fields': ['pub_date']})]
    list_display = ('filename', 'inst', 'pub_date')
    list_filter = ['pub_date']

admin.site.register(ImageData, ImageDataAdmin)



"""Customizes the ``jwql`` app settings.

** CURRENTLY NOT IN USE **

Optionally defines an ``AppConfig`` class that can be called in
``INSTALLED_APPS`` in settings.py to configure the web app.

Authors
-------

    - Lauren Chambers

Use
---

    This module is called in ``settings.py`` as such:
    ::
        INSTALLED_APPS = ['apps.jwql.PlotsExampleConfig',
        ...
        ]

References
----------
    For more information please see:
        ``https://docs.djangoproject.com/en/2.0/ref/applications/``
"""

from django.apps import AppConfig


class PlotsExampleConfig(AppConfig):
    name = 'jwql'



#! /usr/bin/env python
''' Auxiliary module to populate database

    This module was used throughout development to populate the database. Since
    the EDB had no valid data during implementation we had to download data elsewhere.
    The downloaded data is in .CSV format and can easily be read by the program.
    After import and sorting the process_file function extracts the useful part and
    pushes it to the auxiliary database. This function can be implemented in the
    final cron job.

Authors
-------

    - Daniel Kühbacher

Use
---
    make sure "directory" points to a folder where useable 15min-samples are storedself.
    make sure you already ran .utils/sql_interface.py in order to create a empty database
    with prepared tables.
    Run the module form the command line.

Notes
-----
    For developement only
'''

import jwql.instrument_monitors.miri_monitors.data_trending.utils.mnemonics as mn
import jwql.instrument_monitors.miri_monitors.data_trending.utils.sql_interface as sql
import jwql.instrument_monitors.miri_monitors.data_trending.utils.csv_to_AstropyTable as apt
from jwql.instrument_monitors.miri_monitors.data_trending.utils.process_data import once_a_day_routine
from jwql.utils.utils import get_config, filename_parser

import statistics
import os
import glob

#set _location_ variable
__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))

#point to the directory where your files are located!
directory = os.path.join(get_config()['outputs'], 'miri_data_trending', 'trainings_data_15min', '*.CSV')
paths = glob.glob(directory)

def process_file(conn, path):
    '''Parse CSV file, process data within and put to DB

    Parameters
    ----------
    conn : DBobject
        Connection object to auxiliary database
    path : str
        defines file to read
    '''

    #import mnemonic data and append dict to variable below
    m_raw_data = apt.mnemonics(path)

    #process raw data with once a day routine
    processed_data = once_a_day_routine(m_raw_data)

    #push extracted and filtered data to temporary database
    for key, value in processed_data.items():

        #abbreviate data table
        m = m_raw_data.mnemonic(key)

        if key == "SE_ZIMIRICEA":
            length = len(value)
            mean = statistics.mean(value)
            deviation = statistics.stdev(value)
            dataset = (float(m.meta['start']), float(m.meta['end']), length, mean, deviation)
            sql.add_data(conn, "SE_ZIMIRICEA_IDLE", dataset)

        elif key == "IMIR_HK_ICE_SEC_VOLT4":
            length = len(value)
            mean = statistics.mean(value)
            deviation = statistics.stdev(value)
            dataset = (float(m.meta['start']), float(m.meta['end']), length, mean, deviation)
            sql.add_data(conn, "IMIR_HK_ICE_SEC_VOLT4_IDLE", dataset)

        else:
            length = len(value)
            mean = statistics.mean(value)
            deviation = statistics.stdev(value)
            dataset = (float(m.meta['start']), float(m.meta['end']), length, mean, deviation)
            sql.add_data(conn, key, dataset)

def main():
    #generate paths
    DATABASE_LOCATION = os.path.join(get_config()['jwql_dir'], 'database')
    DATABASE_FILE = os.path.join(DATABASE_LOCATION, 'miri_database.db')

    #connect to temporary database
    conn = sql.create_connection(DATABASE_FILE)

    #process every csv file in directory folder
    for path in paths:
        process_file(conn, path)

    #close connection
    sql.close_connection(conn)
    print("done")

if __name__ == "__main__":
    main()



#! /usr/bin/env python

"""
This module is intended to be a template to aid in creating new
monitoring scripts and to demonstrate how to format them to fully
utilize the ``jwql`` framework.

Each monitoring script must be executable from the command line (i.e.
have a ``if '__name__' == '__main__' section), as well as have a "main"
function that calls all other functions, methods, or modules (i.e.
the entirety of the code is executed within the scope of the main
function), as shown in this example.

Users may utilize the ``jwql`` framework functions for logging,
setting permissions, parsing filenames, etc. (See related ``import``s).

Authors
-------

    - Catherine Martlin
    - Matthew Bourque

Use
---

    This module can be executed from the command line:
    ::

        python monitor_template.py

    Alternatively, it can be called from a python environment via the
    following import statements:
    ::

      from monitor_template import main_monitor_function
      from monitor_template import secondary_function

Dependencies
------------

    The user must have a configuration file named ``config.json``
    placed in the ``utils`` directory.

Notes
-----

    Any monitoring script written for ``jwql`` must adhere to the
    ``jwql`` style guide located at:
    https://github.com/spacetelescope/jwql/blob/master/style_guide/style_guide.md
"""

import os
import logging

from astroquery.mast import Mast
from jwst import datamodels
from bokeh.charts import Donut
from bokeh.embed import components

# Functions for logging
from jwql.logging.logging_functions import configure_logging
from jwql.logging.logging_functions import log_info
from jwql.logging.logging_functions import log_fail

# Function for setting permissions of files/directories
from jwql.permissions.permissions import set_permissions

# Function for parsing filenames
from jwql.utils.utils import filename_parser

# Objects for hard-coded information
from jwql.utils.utils import get_config
from jwql.utils.constants import JWST_DATAPRODUCTS, JWST_INSTRUMENT_NAMES


@log_fail
@log_info
def monitor_template_main():
    """ The main function of the ``monitor_template`` module."""

    # Example of logging
    my_variable = 'foo'
    logging.info('Some useful information: {}'.format(my_variable))

    # Example of querying for a dataset via MAST API
    service = "Mast.Jwst.Filtered.Niriss"
    params = {"columns": "filename",
              "filters": [{"paramName": "filter",
                          "values": ['F430M']}]}
    response = Mast.service_request_async(service, params)
    result = response[0].json()['data']
    filename_of_interest = result[0]['filename']  # jw00304002001_02102_00001_nis_uncal.fits

    # Example of parsing a filename
    filename_dict = filename_parser(filename_of_interest)
    # Contents of filename_dict:
    #     {'program_id': '00304',
    #      'observation': '002',
    #      'visit': '001',
    #      'visit_group': '02',
    #      'parallel_seq_id': '1',
    #      'activity': '02',
    #      'exposure_id': '00001',
    #      'detector': 'nis',
    #      'suffix': 'uncal'}

    # Example of locating a dataset in the filesystem
    filesystem = get_config()['filesystem']
    dataset = os.path.join(filesystem, 'jw{}'.format(filename_dict['program_id']),
                           filename_of_interest)

    # Example of reading in dataset using jwst.datamodels
    im = datamodels.open(dataset)
    # Now have access to:
    #     im.data  # Data array
    #     im.err  # ERR array
    #     im.meta  # Metadata such as header keywords

    # Example of saving a file and setting permissions
    im.save('some_filename.fits')
    set_permissions('some_filename.fits')

    # Example of creating and exporting a Bokeh plot
    plt = Donut(im.data, plot_width=600, plot_height=600)
    plt.sizing_mode = 'stretch_both'  # Necessary for responsive sizing on web app
    script, div = components(plt)

    plot_output_dir = get_config()['outputs']
    div_outfile = os.path.join(plot_output_dir, 'monitor_name',
                               filename_of_interest + "_component.html")
    script_outfile = os.path.join(plot_output_dir, 'monitor_name',
                                  filename_of_interest + "_component.js")

    for outfile, component in zip([div_outfile, script_outfile], [div, script]):
        with open(outfile, 'w') as f:
            f.write(component)
            f.close()
        set_permissions(outfile)

    # Perform any other necessary code
    well_named_variable = "Function does something."
    result_of_second_function = second_function(well_named_variable)


def second_function(input_value):
    """ This is your axiliary function; you may have many of these.

    Parameters
    ----------
    input_value : str
        Some value to modify in the function.

    Returns
    -------
    useful_result : str
        The result of modifying the input value.
    """

    # Begin logging:
    logging.info(" ")
    logging.info("The auxiliary function has started running.")

    # Example function:
    useful_result = input_value + " The other function did something, too."

    logging.info("The auxiliary function is returning: ")
    logging.info(useful_result)
    logging.info(" ")

    return useful_result


if __name__ == '__main__':

    # Configure logging
    module = os.path.basename(__file__).strip('.py')
    configure_logging(module)

    # Call the main function
    monitor_template_main()



#! /usr/bin/env python

"""Reset all tables in the ``jwqldb`` database.

Authors
-------

    - Matthew Bourque

Use
---

    This script is intended to be used in the command line:
    ::

        python reset_database.py

Dependencies
------------

    Users must have a ``config.json`` configuration file with a proper
    ``connection_string`` key that points to the ``jwqldb`` database.
    The ``connection_string`` format is
    ``postgresql+psycopg2://user:password@host:port/database``.
"""

from jwql.database.database_interface import base
from jwql.utils.utils import get_config


if __name__ == '__main__':

    connection_string = get_config()['connection_string']
    server_type = connection_string.split('@')[-1][0]

    assert server_type != 'p', 'Cannot reset production database!'

    prompt = ('About to reset all tables for database instance {}. Do you '
              'wish to proceed? (y/n)\n'.format(connection_string))
    response = input(prompt)

    if response.lower() == 'y':
        base.metadata.drop_all()
        base.metadata.create_all()
        print('\nDatabase instance {} has been reset'.format(connection_string))



#! /usr/bin/env python
"""Permissions module for managing file permissions for ``jwql``.

This module provides ``jwql`` with functions to inspect and set file
permissions.

The module takes as input a path to a file or directory, checks whether
the owner of the file is the ``jwql`` admin account, and if so, (1) set
the permissions appropriately, and (2) set the group membership
appropriately.

Authors
-------

    - Johannes Sahlmann

Use
---

    This module can be imported and used with

    ::

        from jwql.permissions import permissions
        permissions.set_permissions(pathname)

    Required arguments:

    ``pathname`` - Directory or file for which the default permissions
    should be set

Notes
-----

    Permissions are set and read using the stat module, see
    https://docs.python.org/3/library/stat.html

    Below is a list with the relevant stat attribute names, integer,
    octal, and string representations.

    ::

        stat_key stat_mode stat_mode_octal stat_mode_string
        -------- --------- --------------- ----------------
        S_IFPORT         0             0o0       ?---------
        S_IFDOOR         0             0o0       ?---------
         S_IXOTH         1             0o1       ?--------x
         S_IWOTH         2             0o2       ?-------w-
         S_IROTH         4             0o4       ?------r--
         S_IRWXO         7             0o7       ?------rwx
         S_IXGRP         8            0o10       ?-----x---
         S_IWGRP        16            0o20       ?----w----
         S_IRGRP        32            0o40       ?---r-----
         S_IRWXG        56            0o70       ?---rwx---
         S_IEXEC        64           0o100       ?--x------
         S_IXUSR        64           0o100       ?--x------
         S_IWUSR       128           0o200       ?-w-------
        S_IWRITE       128           0o200       ?-w-------
         S_IREAD       256           0o400       ?r--------
         S_IRUSR       256           0o400       ?r--------
         S_IRWXU       448           0o700       ?rwx------
         S_ISVTX       512          0o1000       ?--------T
         S_ISGID      1024          0o2000       ?-----S---
         S_ISUID      2048          0o4000       ?--S------
         S_IFIFO      4096             0o0       p---------
         S_IFCHR      8192             0o0       c---------
         S_IFDIR     16384             0o0       d---------
         S_IFBLK     24576             0o0       b---------
         S_IFREG     32768             0o0       ----------
         S_IFLNK     40960             0o0       l---------
        S_IFSOCK     49152             0o0       s---------
         S_IFWHT     57344             0o0       w---------
"""

import grp
import os
import pwd
import stat

# owner and group names for JWQL project
DEFAULT_OWNER = 'jwqladm'
DEFAULT_GROUP = 'jwql_dev'

# set the default mode for DEFAULT_OWNER
DEFAULT_MODE = stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP # equivalent to '?rwxr-x---'

def get_group_string(pathname):
    """Return the group of ``pathname`` in string representation.

    Parameters
    ----------
    pathname : str
        Directory or file to be inspected.

    Returns
    -------
    group_name : str
        String representation of the group.
    """
    file_statinfo = os.stat(pathname)
    groupinfo = grp.getgrgid(file_statinfo.st_gid)
    group_name = groupinfo.gr_name
    return group_name


def get_owner_string(pathname):
    """Return the owner of ``pathname`` in string representation.

    Parameters
    ----------
    pathname : str
        Directory or file to be inspected

    Returns
    -------
    owner_name : str
        String representation of the owner.
    """
    file_statinfo = os.stat(pathname)
    ownerinfo = pwd.getpwuid(file_statinfo.st_uid)
    owner_name = ownerinfo.pw_name
    return owner_name


def has_permissions(pathname, owner=DEFAULT_OWNER, mode=DEFAULT_MODE, group=DEFAULT_GROUP):
    """Return boolean indicating whether ``pathname`` has the specified
    owner, permission, and group scheme.

    Parameters
    ----------
    pathname : str
        Directory or file to be inspected
    owner : str
        String representation of the owner
    mode : int
        Integer representation of the permission mode, compatible
        with ``os.stat`` output
    group : str
        String representation of the group

    Returns
    -------
    boolean : bool
    """
    verify_path(pathname)
    file_statinfo = os.stat(pathname)
    groupinfo = grp.getgrgid(file_statinfo.st_gid)

    # complement mode depending on whether input is file or directory
    if os.path.isfile(pathname):
        mode = mode | stat.S_IFREG
    elif os.path.isdir(pathname):
        mode = mode | stat.S_IFDIR

    if (get_owner_string(pathname) != owner) or (file_statinfo.st_mode != mode)\
            or (groupinfo.gr_name != group):
        return False

    return True


def set_permissions(pathname, owner=DEFAULT_OWNER, mode=DEFAULT_MODE, group=DEFAULT_GROUP, verbose=False):
    """Set mode and group of the file/directory identfied by
    ``pathname``, if and only if it is owned by ``owner``.

    Parameters
    ----------
    pathname : str
        Directory or file to be inspected
    owner : str
        String representation of the owner
    mode : int
        Integer representation of the permission mode, compatible with
        ``os.stat`` output
    group : str
        String representation of the group
    verbose : bool
        Boolean indicating whether verbose output is requested
    """
    if verbose:
        print('\nBefore:')
        show_permissions(pathname)

    if not has_permissions(pathname):
        if get_owner_string(pathname) == owner:
            os.chmod(pathname, mode)
            # change group but not owner
            os.chown(pathname, -1, grp.getgrnam(group).gr_gid)

    if verbose:
        print('After:')
        show_permissions(pathname)


def show_permissions(pathname):
    """Verbose output showing group, user, and permission information
    for a directory or file.

    Parameters
    ----------
    pathname : str
        Directory or file to be inspected
    """
    verify_path(pathname)
    file_statinfo = os.stat(pathname)
    ownerinfo = pwd.getpwuid(file_statinfo.st_uid)
    groupinfo = grp.getgrgid(file_statinfo.st_gid)

    if os.path.isdir(pathname):
        info_string = 'directory'
    elif os.path.isfile(pathname):
        info_string = 'file'

    print('Inspecting {}: {}'.format(info_string, pathname))
    print('group {} = {}'.format(file_statinfo.st_gid, groupinfo.gr_name))
    print('owner {} = {}'.format(file_statinfo.st_uid, ownerinfo.pw_name))
    print('mode {} = {}'.format(file_statinfo.st_mode, stat.filemode(file_statinfo.st_mode)))
    print('')


def verify_path(pathname):
    """Verify that pathname is either a directory or a file. If not, an
    error is raised.

    Parameters
    ----------
    pathname : str
        Directory or file to be inspected
    """
    if (not os.path.isdir(pathname)) and (not os.path.isfile(pathname)):
        raise NotImplementedError('{} is not a valid path or filename'.format(pathname))



#!/usr/bin/env python
"""Tests for the ``utils`` module.

Authors
-------

    - Lauren Chambers
    - Matthew Bourque

Use
---

    These tests can be run via the command line (omit the -s to
    suppress verbose output to stdout):

    ::

        pytest -s test_utils.py
"""

import os
from pathlib import Path
import pytest

from jwql.utils.utils import copy_files, get_config, filename_parser, filesystem_path


FILENAME_PARSER_TEST_DATA = [

    # Test full path
    ('/test/dir/to/the/file/jw90002/jw90002001001_02102_00001_nis_rateints.fits',
     {'activity': '02',
      'detector': 'nis',
      'exposure_id': '00001',
      'filename_type': 'stage_1_and_2',
      'instrument': 'niriss',
      'observation': '001',
      'parallel_seq_id': '1',
      'program_id': '90002',
      'suffix': 'rateints',
      'visit': '001',
      'visit_group': '02'}),

    # Test full stage 1 and 2 filename
    ('jw00327001001_02101_00002_nrca1_rate.fits',
     {'activity': '01',
      'detector': 'nrca1',
      'exposure_id': '00002',
      'filename_type': 'stage_1_and_2',
      'instrument': 'nircam',
      'observation': '001',
      'parallel_seq_id': '1',
      'program_id': '00327',
      'suffix': 'rate',
      'visit': '001',
      'visit_group': '02'}),

    # Test root stage 1 and 2 filename
    ('jw00327001001_02101_00002_nrca1',
     {'activity': '01',
      'detector': 'nrca1',
      'exposure_id': '00002',
      'filename_type': 'stage_1_and_2',
      'instrument': 'nircam',
      'observation': '001',
      'parallel_seq_id': '1',
      'program_id': '00327',
      'visit': '001',
      'visit_group': '02'}),

    # Test full stage 2c filename
    ('jw94015002002_02108_00001_mirimage_o002_crf.fits',
     {'ac_id': 'o002',
      'activity': '08',
      'detector': 'mirimage',
      'exposure_id': '00001',
      'filename_type': 'stage_2c',
      'instrument': 'miri',
      'observation': '002',
      'parallel_seq_id': '1',
      'program_id': '94015',
      'suffix': 'crf',
      'visit': '002',
      'visit_group': '02'}),

    # Test root stage 2c filename
    ('jw90001001003_02101_00001_nis_o001',
     {'ac_id': 'o001',
      'activity': '01',
      'detector': 'nis',
      'exposure_id': '00001',
      'filename_type': 'stage_2c',
      'instrument': 'niriss',
      'observation': '001',
      'parallel_seq_id': '1',
      'program_id': '90001',
      'visit': '003',
      'visit_group': '02'}),

    # Test full stage 3 filename with target_id
    ('jw80600-o009_t001_miri_f1130w_i2d.fits',
     {'ac_id': 'o009',
      'filename_type': 'stage_3_target_id',
      'instrument': 'miri',
      'optical_elements': 'f1130w',
      'program_id': '80600',
      'suffix': 'i2d',
      'target_id': 't001'}),

    # Test full stage 3 filename with target_id and different ac_id
    ('jw80600-c0001_t001_miri_f1130w_i2d.fits',
     {'ac_id': 'c0001',
      'filename_type': 'stage_3_target_id',
      'instrument': 'miri',
      'optical_elements': 'f1130w',
      'program_id': '80600',
      'suffix': 'i2d',
      'target_id': 't001'}),

    # Test full stage 3 filename with source_id
    ('jw80600-o009_s00001_miri_f1130w_i2d.fits',
     {'ac_id': 'o009',
      'filename_type': 'stage_3_source_id',
      'instrument': 'miri',
      'optical_elements': 'f1130w',
      'program_id': '80600',
      'source_id': 's00001',
      'suffix': 'i2d'}),

    # Test stage 3 filename with target_id and epoch
    ('jw80600-o009_t001-epoch1_miri_f1130w_i2d.fits',
     {'ac_id': 'o009',
      'filename_type': 'stage_3_target_id_epoch',
      'instrument': 'miri',
      'epoch': '1',
      'optical_elements': 'f1130w',
      'program_id': '80600',
      'suffix': 'i2d',
      'target_id': 't001'}),

    # Test stage 3 filename with source_id and epoch
    ('jw80600-o009_s00001-epoch1_miri_f1130w_i2d.fits',
     {'ac_id': 'o009',
      'filename_type': 'stage_3_source_id_epoch',
      'instrument': 'miri',
      'epoch': '1',
      'optical_elements': 'f1130w',
      'program_id': '80600',
      'source_id': 's00001',
      'suffix': 'i2d'}),

    # Test root stage 3 filename with target_id
    ('jw80600-o009_t001_miri_f1130w',
     {'ac_id': 'o009',
      'filename_type': 'stage_3_target_id',
      'instrument': 'miri',
      'optical_elements': 'f1130w',
      'program_id': '80600',
      'target_id': 't001'}),

    # Test root stage 3 filename with source_id
    ('jw80600-o009_s00001_miri_f1130w',
     {'ac_id': 'o009',
      'filename_type': 'stage_3_source_id',
      'instrument': 'miri',
      'optical_elements': 'f1130w',
      'program_id': '80600',
      'source_id': 's00001'}),

    # Test full time series filename
    ('jw00733003001_02101_00002-seg001_nrs1_rate.fits',
     {'activity': '01',
      'detector': 'nrs1',
      'exposure_id': '00002',
      'filename_type': 'time_series',
      'instrument': 'nirspec',
      'observation': '003',
      'parallel_seq_id': '1',
      'program_id': '00733',
      'segment': '001',
      'suffix': 'rate',
      'visit': '001',
      'visit_group': '02'}),

    # Test root time series filename
    ('jw00733003001_02101_00002-seg001_nrs1',
     {'activity': '01',
      'detector': 'nrs1',
      'exposure_id': '00002',
      'filename_type': 'time_series',
      'instrument': 'nirspec',
      'observation': '003',
      'parallel_seq_id': '1',
      'program_id': '00733',
      'segment': '001',
      'visit': '001',
      'visit_group': '02'}),

    # Test full guider ID filename
    ('jw00729011001_gs-id_1_image_cal.fits',
     {'date_time': None,
      'filename_type': 'guider',
      'guide_star_attempt_id': '1',
      'guider_mode': 'id',
      'instrument': 'fgs',
      'observation': '011',
      'program_id': '00729',
      'suffix': 'image_cal',
      'visit': '001'}),

    # Test root guider ID filename
    ('jw00327001001_gs-id_2',
     {'date_time': None,
      'filename_type': 'guider',
      'guide_star_attempt_id': '2',
      'guider_mode': 'id',
      'instrument': 'fgs',
      'observation': '001',
      'program_id': '00327',
      'visit': '001'}),

    # Test full guider non-ID filename
    ('jw86600048001_gs-fg_2016018175411_stream.fits',
     {'date_time': '2016018175411',
      'filename_type': 'guider',
      'guide_star_attempt_id': None,
      'guider_mode': 'fg',
      'instrument': 'fgs',
      'observation': '048',
      'program_id': '86600',
      'suffix': 'stream',
      'visit': '001'}),

    # Test root guider non-ID filename
    ('jw00729011001_gs-acq2_2019155024808',
     {'date_time': '2019155024808',
      'filename_type': 'guider',
      'guide_star_attempt_id': None,
      'guider_mode': 'acq2',
      'instrument': 'fgs',
      'observation': '011',
      'program_id': '00729',
      'visit': '001'})

]


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_copy_files():
    """Test that files are copied successfully"""

    # Create an example file to be copied
    data_dir = os.path.dirname(__file__)
    file_to_copy = 'file.txt'
    original_file = os.path.join(data_dir, file_to_copy)
    Path(original_file).touch()
    assert os.path.exists(original_file), 'Failed to create original test file.'

    # Make a copy one level up
    new_location = os.path.abspath(os.path.join(data_dir, '../'))
    copied_file = os.path.join(new_location, file_to_copy)

    # Copy the file
    success, failure = copy_files([original_file], new_location)
    assert success == [copied_file]
    assert os.path.isfile(copied_file)

    # Remove the copy
    os.remove(original_file)
    os.remove(copied_file)


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_get_config():
    """Assert that the ``get_config`` function successfully creates a
    dictionary.
    """
    settings = get_config()
    assert isinstance(settings, dict)


@pytest.mark.parametrize('filename, solution', FILENAME_PARSER_TEST_DATA)
def test_filename_parser(filename, solution):
    """Generate a dictionary with parameters from a JWST filename.
    Assert that the dictionary matches what is expected.

    Parameters
    ----------
    filename : str
        The filename to test (e.g. ``jw00327001001_02101_00002_nrca1_rate.fits``)
    solution : dict
        A dictionary of the expected result
    """

    assert filename_parser(filename) == solution


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_filename_parser_whole_filesystem():
    """Test the filename_parser on all files currently in the filesystem."""
    # Get all files
    filesystem_dir = get_config()['filesystem']
    all_files = []
    for dir_name, _, file_list in os.walk(filesystem_dir):
        for file in file_list:
            if file.endswith('.fits'):
                all_files.append(os.path.join(dir_name, file))

    # Run the filename_parser on all files
    bad_filenames = []
    for filepath in all_files:
        try:
            filename_parser(filepath)
        except ValueError:
            bad_filenames.append(os.path.basename(filepath))

    # Determine if the test failed
    fail = bad_filenames != []
    failure_msg = '{} files could not be successfully parsed: \n - {}'.\
        format(len(bad_filenames), '\n - '.join(bad_filenames))

    # Check which ones failed
    assert not fail, failure_msg


def test_filename_parser_nonJWST():
    """Attempt to generate a file parameter dictionary from a file
    that is not formatted in the JWST naming convention. Ensure the
    appropriate error is raised.
    """
    with pytest.raises(ValueError):
        filename = 'not_a_jwst_file.fits'
        filename_parser(filename)


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_filesystem_path():
    """Test that a file's location in the filesystem is returned"""

    filename = 'jw96003001001_02201_00001_nrca1_dark.fits'
    check = filesystem_path(filename)
    location = os.path.join(get_config()['filesystem'], 'jw96003', filename)

    assert check == location



#! /usr/bin/env python

"""Tests for the ``permissions`` module.

Authors
-------

    - Johannes Sahlmann


Use
---

    These tests can be run via the command line (omit the ``-s`` to
    suppress verbose output to ``stdout``):

    ::

        pytest -s test_permissions.py
"""

import grp
import os
import pytest

from jwql.utils.permissions import set_permissions, has_permissions, \
    get_owner_string, get_group_string

# directory to be created and populated during tests running
TEST_DIRECTORY = os.path.join(os.environ['HOME'], 'permission_test')


@pytest.fixture(scope="module")
def test_directory(test_dir=TEST_DIRECTORY):
    """Create a test directory for permission management.

    Parameters
    ----------
    test_dir : str
        Path to directory used for testing

    Yields
    -------
    test_dir : str
        Path to directory used for testing
    """
    os.mkdir(test_dir)  # creates directory with default mode=511

    yield test_dir
    if os.path.isdir(test_dir):
        os.remove(test_dir)


def test_directory_permissions(test_directory):
    """Create a directory with the standard permissions
    ``('-rw-r--r--')``.

    Set the default permissions defined in ``permissions.py``. Assert
    that these were set correctly.

    Parameters
    ----------
    test_directory : str
        Path of directory used for testing
    """
    # Get owner and group on the current system.This allows to implement the tests
    # independently from the user.
    owner = get_owner_string(test_directory)
    group = get_group_string(test_directory)

    set_permissions(test_directory, owner=owner, group=group)
    assert has_permissions(test_directory, owner=owner, group=group)


@pytest.fixture()
def test_file(test_dir=TEST_DIRECTORY):
    """Create a test file for permission management.

    Parameters
    ----------
    test_dir : str
        Path to directory used for testing

    Yields
    -------
    filename : str
        Path of file used for testing
    """
    if not os.path.isdir(test_dir):
        os.mkdir(test_dir)

    filename = os.path.join(test_dir, 'permission_test.txt')
    with open(filename, 'w') as filestream:
        filestream.write('jwql permission test')
    yield filename
    if os.path.isfile(filename):
        os.remove(filename)
    if os.path.isdir(test_dir):
        os.rmdir(test_dir)


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_file_group(test_file):
    """Create a file with the standard permissions ``('-rw-r--r--')``
    and default group.

    Modify the group and set the default permissions defined in
    ``permissions.py``.  Assert that both group and permissions were
    set correctly.

    Parameters
    ----------
    test_file : str
        Path of file used for testing
    """
    # Get owner and group on the current system.
    owner = get_owner_string(test_file)
    group = get_group_string(test_file)

    # attempt to retrieve a group name different from default
    group_index = 0
    test_group = grp.getgrgid(os.getgroups()[group_index]).gr_name

    set_permissions(test_file, group=test_group, owner=owner)
    assert has_permissions(test_file, group=test_group, owner=owner)

    # return to default group
    set_permissions(test_file, owner=owner, group=group)
    assert has_permissions(test_file, owner=owner, group=group)


def test_file_permissions(test_file):
    """Create a file with the standard permissions ``('-rw-r--r--')``.

    Set the default permissions defined in ``permissions.py``. Assert
    that these were set correctly.

    Parameters
    ----------
    test_file : str
        Path of file used for testing
    """
    # Get owner and group on the current system.
    owner = get_owner_string(test_file)
    group = get_group_string(test_file)

    set_permissions(test_file, owner=owner, group=group)
    assert has_permissions(test_file, owner=owner, group=group)



"""Write unit tests for the ``jwql`` app.

** CURRENTLY NOT IN USE **

Use pytest to run unit tests on portions of the code.

References
----------
For more information please see:
    https://docs.djangoproject.com/en/2.0/internals/contributing/writing-code/unit-tests/

"""

from django.test import TestCase

# Create your tests here.



#! /usr/bin/env python

"""Tests for the ``pipeline_tools`` module.

Authors
-------

    - Bryan Hilbert

Use
---

    These tests can be run via the command line (omit the ``-s`` to
    suppress verbose output to stdout):
    ::

        pytest -s test_pipeline_tools.py
"""

from collections import OrderedDict
import os
import pytest

import numpy as np

from jwql.instrument_monitors import pipeline_tools
from jwql.utils.utils import get_config


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_completed_pipeline_steps():
    """Test that the list of completed pipeline steps for a file is
    correct

    Parameters
    ----------
    filename : str
        File to be checked
    """

    filename = os.path.join(get_config()['filesystem'], 'jw00312', 'jw00312002001_02102_00001_nrcb4_rateints.fits')
    completed_steps = pipeline_tools.completed_pipeline_steps(filename)
    true_completed = OrderedDict([('group_scale', False),
                                  ('dq_init', True),
                                  ('saturation', True),
                                  ('ipc', False),
                                  ('refpix', True),
                                  ('superbias', True),
                                  ('persistence', True),
                                  ('dark_current', True),
                                  ('linearity', True),
                                  ('firstframe', False),
                                  ('lastframe', False),
                                  ('rscd', False),
                                  ('jump', True),
                                  ('rate', True)])

    # Only test steps that have a value of True
    completed_steps = OrderedDict((k, v) for k, v in completed_steps.items() if v is True)
    true_completed = OrderedDict((k, v) for k, v in true_completed.items() if v is True)

    assert completed_steps == true_completed


def test_get_pipeline_steps():
    """Test that the proper pipeline steps are returned for an
    instrument
    """

    # FGS, NIRCam, and NIRISS
    instruments = ['fgs', 'nircam', 'niriss']
    for instrument in instruments:
        req_steps = pipeline_tools.get_pipeline_steps(instrument)
        steps = ['dq_init', 'saturation', 'superbias', 'refpix', 'linearity',
                 'persistence', 'dark_current', 'jump', 'rate']
        not_required = ['group_scale', 'ipc', 'firstframe', 'lastframe', 'rscd']
        steps_dict = OrderedDict({})
        for step in steps:
            steps_dict[step] = True
        for step in not_required:
            steps_dict[step] = False
        # Only test steps that have a value of True
        req_steps = OrderedDict((k, v) for k, v in req_steps.items() if v is True)
        steps_dict = OrderedDict((k, v) for k, v in steps_dict.items() if v is True)
        assert req_steps == steps_dict

    # NIRSpec
    nrs_req_steps = pipeline_tools.get_pipeline_steps('nirspec')
    nrs_steps = ['group_scale', 'dq_init', 'saturation', 'superbias', 'refpix', 'linearity',
                 'dark_current', 'jump', 'rate']
    not_required = ['ipc', 'persistence', 'firstframe', 'lastframe', 'rscd']
    nrs_dict = OrderedDict({})
    for step in nrs_steps:
        nrs_dict[step] = True
    for step in not_required:
        nrs_dict[step] = False
    # Only test steps that have a value of True
    nrs_req_steps = OrderedDict((k, v) for k, v in nrs_req_steps.items() if v is True)
    nrs_dict = OrderedDict((k, v) for k, v in nrs_dict.items() if v is True)
    assert nrs_req_steps == nrs_dict

    # MIRI
    miri_req_steps = pipeline_tools.get_pipeline_steps('miri')
    miri_steps = ['dq_init', 'saturation', 'firstframe', 'lastframe',
                  'linearity', 'rscd', 'dark_current', 'refpix', 'jump', 'rate']
    not_required = ['group_scale', 'ipc', 'superbias', 'persistence']
    miri_dict = OrderedDict({})
    for step in miri_steps:
        miri_dict[step] = True
    for step in not_required:
        miri_dict[step] = False
    # Only test steps that have a value of True
    miri_req_steps = OrderedDict((k, v) for k, v in miri_req_steps.items() if v is True)
    miri_dict = OrderedDict((k, v) for k, v in miri_dict.items() if v is True)
    assert miri_req_steps == miri_dict


@pytest.mark.skipif(os.path.expanduser('~') == '/home/jenkins',
                    reason='Requires access to central storage.')
def test_image_stack():
    """Test stacking of slope images"""

    directory = os.path.join(get_config()['test_dir'], 'dark_monitor')
    files = [os.path.join(directory, 'test_image_{}.fits'.format(str(i+1))) for i in range(3)]

    image_stack, exptimes = pipeline_tools.image_stack(files)
    truth = np.zeros((3, 10, 10))
    truth[0, :, :] = 5.
    truth[1, :, :] = 10.
    truth[2, :, :] = 15.

    assert np.all(image_stack == truth)
    assert exptimes == [[10.5], [10.5], [10.5]]


def test_steps_to_run():
    """Test that the dictionaries for steps required and steps completed
    are correctly combined to create a dictionary of pipeline steps to
    be done

    Parameters
    ----------
    filename : str
        File to be checked

    required : OrderedDict
        Dict of all pipeline steps to be run on filename

    already_done : OrderedDict
        Dict of pipeline steps already run on filename
    """

    required = OrderedDict([('group_scale', True),
                            ('dq_init', False),
                            ('saturation', False),
                            ('ipc', False),
                            ('refpix', False),
                            ('superbias', False),
                            ('persistence', True),
                            ('dark_current', True),
                            ('linearity', False),
                            ('firstframe', False),
                            ('lastframe', False),
                            ('rscd', False),
                            ('jump', True),
                            ('rate', True)])
    already_done = OrderedDict([('group_scale', True),
                                ('dq_init', False),
                                ('saturation', False),
                                ('ipc', False),
                                ('refpix', False),
                                ('superbias', False),
                                ('persistence', True),
                                ('dark_current', True),
                                ('linearity', False),
                                ('firstframe', False),
                                ('lastframe', False),
                                ('rscd', False),
                                ('jump', False),
                                ('rate', False)])

    steps_to_run = pipeline_tools.steps_to_run(required, already_done)
    true_steps_to_run = OrderedDict([('group_scale', False),
                                     ('dq_init', False),
                                     ('saturation', False),
                                     ('ipc', False),
                                     ('refpix', False),
                                     ('superbias', False),
                                     ('persistence', False),
                                     ('dark_current', False),
                                     ('linearity', False),
                                     ('firstframe', False),
                                     ('lastframe', False),
                                     ('rscd', False),
                                     ('jump', True),
                                     ('rate', True)])

    assert steps_to_run == true_steps_to_run
